#!/usr/bin/env python
# coding: utf-8

"""
Reproduction of Fig. 2 (b and c) of Spreizer et al. 2019 in Brian: 

Spreizer S, Aertsen A, Kumar A. From space to time: Spatial inhomogeneities lead to the emergence of spatiotemporal sequences in spiking neuronal networks. PLOS Computational Biology. 2019;15(10):e1007432. doi:10.1371/journal.pcbi.1007432

- Requires the ``noise`` package for setting up the Perlin landscape with freshly generated noise. Otherwise,
  the landscapes for different network sizes (scaled down by a factor of 1 to 5) can be
  loaded from the provided files.

This code was written by Arash Golmohammadi, 2022. Modified plotting code by Marcel Stimberg, 2023.
"""
import os
import numpy as np
import brian2 as b2
from brian2.units import ms, pA, mV

import matplotlib.pyplot as plt
from matplotlib import animation

b2.seed(8)


def round_to_even(gs, scaler):
    """
    rounds the network size to an even number.
    """
    rounded = round(gs / scaler)
    if rounded % 2:
        rounded += 1
    return int(rounded)


# ---------------- CONFIGURATIONS ---------------- #

# NETWORK TOPOLOGY
SCALE_DOWN = 5  # For faster computation I recommend running with SCALE_DOWN 2 or 3
GRID_SIZE = round_to_even(100, SCALE_DOWN)
NCONN = round_to_even(1000, SCALE_DOWN**2)

# CELL CONFIGS
THR, REF, RESET = -55 * mV, 2 * ms, -70 * mV

# PROFILE
THETA, KAPPA, GAP = 3 / SCALE_DOWN, 4, 3

# ANISOTROPY
R, SCALE = 1.0, 4  # perlin noise
PHI_H = np.pi / 6  # homog. angle

# WARMUP
DUR_WU, STD_WU = 500 * ms, 500 * pA

# BACKGROUND
MU, SIGMA = 700 * pA, 100 * pA


EQ_NRN = """
    dv/dt = (E-v)/tau_m + (noise + I_syn)/C : volt (unless refractory)
    I_syn :  amp
    noise  = mu + sigma*sqrt(noise_dt)*xi_pop: amp
    
    mu               : amp (shared)
    sigma            : amp (shared)
    noise_dt = 1*ms  : second (shared)
    C = 250*pF       : farad (shared)
    tau_m = 10*ms      : second (shared)
    E = -70*mV       : volt (shared)
"""

EQ_SYN = f"""
    dg/dt = (-g+h) / tau_s : 1 (clock-driven)
    dh/dt = -h / tau_s     : 1 (clock-driven)
    I_syn_post = J*g  : amp  (summed)
    J = -{SCALE_DOWN}*10*pA           : amp (shared)
    tau_s = 5*ms           : second (shared)
"""

ON_PRE = "h += exp(1)"  # to match NEST's normalization


# ---------------- UTILITIES ---------------- #
def coord2idx(coords, pop):
    """Converts array of coordinates to indices."""
    coords = np.asarray(coords).reshape(-1, 2)
    idxs = coords[:, 1] * GRID_SIZE + coords[:, 0]
    return idxs


def idx2coord(idxs, pop):
    """Converts array of indices to coordinates"""
    idxs = np.asarray(idxs)
    y, x = np.divmod(idxs, GRID_SIZE)
    coords = np.array([x, y]).T
    return coords


def make_periodic(x, y, grid_size=GRID_SIZE):
    x = x % GRID_SIZE
    y = y % GRID_SIZE
    return x, y


# ---------------- LANDSCAPE GENERATION ---------------- #
def landscape(anisotropy="perlin"):
    """Generates an angular bias for each neuron in the
    network such that no angle is more likely than the
    other. Yet, nearby bias angles may be correleted.

    Note that this is a flattened array."""

    if anisotropy == "perlin":  # requires noise package
        if os.path.exists(f"perlin{SCALE_DOWN}.npy"):
            print("Loading perlin noise from file...")
            phis = np.load(f"perlin{SCALE_DOWN}.npy")
        else:
            print("Generating perlin noise...")
            from noise import pnoise2 as perlin

            x = y = np.linspace(0, SCALE, GRID_SIZE)
            phis = [[perlin(i, j, repeatx=SCALE, repeaty=SCALE) for j in y] for i in x]
            phis = np.concatenate(phis)
            phis = balance_landscape(phis)
            phis -= phis.min()
            phis *= 2 * np.pi / phis.max()
    elif anisotropy == "homog":
        phis = np.ones(GRID_SIZE**2) * PHI_H
    else:  # both random and symmetric have the same landscape
        phis = np.random.uniform(0, 2 * np.pi, size=GRID_SIZE**2)

    return phis


def balance_landscape(array):
    """Equalizes the histogram."""

    sorted_idx = np.argsort(array)
    gs = int(np.sqrt(len(array)))
    max_val = gs * 2
    idx = int(len(array) // max_val)
    for i, val in enumerate(range(max_val)):
        array[sorted_idx[i * idx : (i + 1) * idx]] = val

    return (array - gs) / gs

# ---------------- NETWORK SETUP ---------------- #
def setup_net(phis):
    I = b2.NeuronGroup(
        N=GRID_SIZE**2,
        name="I",
        model=EQ_NRN,
        refractory=5 * ms,
        threshold=f"v>{THR / mV}*mV",
        reset=f"v={RESET / mV}*mV",
        method="euler",
    )
    I.mu = MU
    I.sigma = SIGMA

    syn_II = b2.Synapses(
        I, I, model=EQ_SYN, on_pre=ON_PRE, method="exact", name="syn_II"
    )

    for s_idx in range(len(I)):
        rel_coords = draw_post(s_idx, phis[s_idx])
        s_coord = idx2coord(s_idx, I)  # check shape
        x, y = (rel_coords + s_coord).T
        x, y = make_periodic(x, y)
        t_coords = np.array([x, y]).T.astype(int)
        t_idxs = coord2idx(t_coords, I)

        syn_II.connect(i=s_idx, j=t_idxs)

    return I, syn_II


def draw_post(idx, phi):
    alpha = np.random.uniform(-np.pi, np.pi, NCONN)
    radius = np.concatenate(
        [
            -np.random.gamma(shape=KAPPA, scale=THETA, size=int(NCONN // 2)),
            +np.random.gamma(shape=KAPPA, scale=THETA, size=NCONN - int(NCONN // 2)),
        ]
    )

    radius[radius < 0] -= GAP
    radius[radius >= 0] += GAP

    x, y = radius * np.cos(alpha), radius * np.sin(alpha)
    x += R * np.cos(phi)
    y += R * np.sin(phi)

    eps = 1e-3
    self_link = x**2 + y**2 < eps**2
    x = x[~self_link]
    y = y[~self_link]

    coords = np.array([x, y]).T
    return np.round(coords).astype(int)


def warmup(net):
    net["I"].mu = 0 * pA
    net["I"].sigma = STD_WU

    net.run(DUR_WU / 2)
    net["I"].mu = MU
    net["I"].sigma = SIGMA
    net.run(DUR_WU / 2)


def simulate(phis, dur=3000 * ms):
    b2.start_scope()
    I, syn_II = setup_net(phis)
    mon = b2.SpikeMonitor(I, record=True)

    net = b2.Network()
    net.add(I)
    net.add(syn_II)
    net.add(mon)

    warmup(net)
    net.run(dur)

    return net, mon

# ---------------- VISUALIZATION ---------------- #
def plot_firing_rate_disto(idxs, ts, ax, name):
    """plots distribution of firing rates over the simulation"""

    T = np.max(ts) - np.min(ts)
    _, rates = np.unique(idxs, return_counts=True)
    rates = rates * 1.0 / T
    ax.hist(
        rates,
        bins=25,
        density=True,
        label=name,
        histtype="step",
        orientation="horizontal",
        lw=2,
    )
    ax.set(
        ylabel="Firing rate [Hz]",
        xlabel="Probability density",
        xscale="log",
        ylim=(0, 64),
    )

def plot_in_deg(in_deg, ax, ax_hist, name):
    """plots distribution as well as the heat map of in-degree from each neuron."""
    # field map & distribution of in-degrees
    m = ax.pcolormesh(in_deg, shading="flat", vmin=0.5 * NCONN, vmax=1.5 * NCONN)
    ax_hist.hist(
        in_deg.flatten(),
        bins=50,
        density=True,
        label=name,
        histtype="step",
        orientation="horizontal",
        lw=2,
    )

    ax_hist.set(ylabel="In-degree", ylim=(0.5 * NCONN, 1.5 * NCONN))

    ax.set(
        aspect="equal",
        xticks=[],
        yticks=[],
        title=name,
    )
    ax_hist.locator_params(axis="x", nbins=3)
    return m

def animator(fig, all_imgs, all_vals):
    """used for making an activity animation"""
    n_frames = len(all_vals[0][0])  # total number of frames

    def animate(frame_id):
        for img, vals in zip(all_imgs, all_vals):
            img.set_array(vals[frame_id])
        return all_imgs

    # Call the animator.
    anim = animation.FuncAnimation(
        fig, animate, frames=n_frames, interval=50, blit=True
    )

    return anim

def plot_animations(all_idxs, all_ts, duration, fig, axs, cax, ss_dur=25):
    """Plots neuronal activity as an animation"""

    ts_bins = np.arange(0, duration / ms + 1, ss_dur) * ms

    field_vals, field_imgs = [], []
    for idxs, ts, ax in zip(all_idxs, all_ts, axs.flat):
        h = np.histogram2d(ts, idxs, bins=[ts_bins, range(GRID_SIZE**2 + 1)])[0]
        field_val = h.reshape(-1, GRID_SIZE, GRID_SIZE) / (float(ss_dur * ms))
        field_img = ax.imshow(field_val[0], vmin=0, vmax=64, origin="lower")
        field_imgs.append(field_img)
        field_vals.append(field_val)
        ax.set(xticks=[], yticks=[])
    
    plt.colorbar(field_img, cax=cax)

    anim = animator(fig, field_imgs, field_vals)
    return anim


# ---------------- MAIN ---------------- #
fig, axs = plt.subplots(
    2,
    6,
    figsize=(15, 6),
    layout="constrained",
    gridspec_kw={"width_ratios": [1, 1, 1, 1, 0.1, 1]},
)
all_idxs, all_ts = [], []

for idx, lscp in enumerate(["symmetric", "random", "perlin", "homog"]):
    print(f"Simulating {lscp} landscape.")

    phis = landscape(lscp)
    net, mon = simulate(phis)
    
    ## Static plots
    # In-degrees
    syn = net["syn_II"]
    in_deg = syn.N_incoming_post.reshape((GRID_SIZE, GRID_SIZE))

    m = plot_in_deg(in_deg, axs[0, idx], axs[0, 5], lscp)

    # Firing rates
    idxs, ts = mon.it
    all_idxs.append(idxs)
    all_ts.append(ts)
    plot_firing_rate_disto(idxs, ts, axs[1, 5], lscp)

# Add colorbar and legend for in-degree plots
plt.colorbar(m, cax=axs[0, 4])
axs[0, -1].legend()

# Plot animations
duration = net.t
anim = plot_animations(all_idxs, all_ts, duration, fig, axs[1, :4], axs[1, 4], ss_dur=25)

plt.show()



#!/usr/bin/env python
"""
Dynamics of a network of sparsely connected inhibitory current-based 
integrate-and-fire neurons. Individual neurons fire irregularly at 
low rate but the network is in an oscillatory global activity regime 
where neurons are weakly synchronized.

Reference:
    "Fast Global Oscillations in Networks of Integrate-and-Fire
    Neurons with Low Firing Rates"
    Nicolas Brunel & Vincent Hakim
    Neural Computation 11, 1621-1671 (1999)
"""

from brian2 import *

N = 5000
Vr = 10*mV
theta = 20*mV
tau = 20*ms
delta = 2*ms
taurefr = 2*ms
duration = .1*second
C = 1000
sparseness = float(C)/N
J = .1*mV
muext = 25*mV
sigmaext = 1*mV

eqs = """
dV/dt = (-V+muext + sigmaext * sqrt(tau) * xi)/tau : volt
"""

group = NeuronGroup(N, eqs, threshold='V>theta',
                    reset='V=Vr', refractory=taurefr, method='euler')
group.V = Vr
conn = Synapses(group, group, on_pre='V += -J', delay=delta)
conn.connect(p=sparseness)
M = SpikeMonitor(group)
LFP = PopulationRateMonitor(group)

run(duration)

subplot(211)
plot(M.t/ms, M.i, '.')
xlim(0, duration/ms)

subplot(212)
plot(LFP.t/ms, LFP.smooth_rate(window='flat', width=0.5*ms)/Hz)
xlim(0, duration/ms)

show()



import pytest
from numpy.testing import assert_equal

from brian2 import *
from brian2.devices.device import reinit_and_delete


@pytest.mark.standalone_compatible
def test_simple_threshold():
    G = NeuronGroup(4, "v : 1", threshold="v > 1")
    G.v = [1.5, 0, 3, -1]
    s_mon = SpikeMonitor(G)
    run(defaultclock.dt)
    assert_equal(s_mon.count, np.array([1, 0, 1, 0]))


@pytest.mark.standalone_compatible
def test_scalar_threshold():
    c = 2
    G = NeuronGroup(4, "", threshold="c > 1")
    s_mon = SpikeMonitor(G)
    run(defaultclock.dt)
    assert_equal(s_mon.count, np.array([1, 1, 1, 1]))


if __name__ == "__main__":
    test_simple_threshold()
    test_scalar_threshold()



'''
Run all the numpy tests using pytest. Exits with error code 1 if a test failed.
'''
import sys

import brian2

if not brian2.test('numpy', test_codegen_independent=False):  # If the test fails, exit with a non-zero error code
    sys.exit(1)



"""
Module defining `CodeString`, a class for a string of code together with
information about its namespace. Only serves as a parent class, its subclasses
`Expression` and `Statements` are the ones that are actually used.
"""

from collections.abc import Hashable

import sympy

from brian2.parsing.sympytools import str_to_sympy, sympy_to_str
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers

__all__ = ["Expression", "Statements"]

logger = get_logger(__name__)


class CodeString(Hashable):
    """
    A class for representing "code strings", i.e. a single Python expression
    or a sequence of Python statements.

    Parameters
    ----------
    code : str
        The code string, may be an expression or a statement(s) (possibly
        multi-line).

    """

    def __init__(self, code):
        self._code = code.strip()

        # : Set of identifiers in the code string
        self.identifiers = get_identifiers(code)

    code = property(lambda self: self._code, doc="The code string")

    def __str__(self):
        return self.code

    def __repr__(self):
        return f"{self.__class__.__name__}({self.code!r})"

    def __eq__(self, other):
        if not isinstance(other, CodeString):
            return NotImplemented
        return self.code == other.code

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.code)


class Statements(CodeString):
    """
    Class for representing statements.

    Parameters
    ----------
    code : str
        The statement or statements. Several statements can be given as a
        multi-line string or separated by semicolons.

    Notes
    -----
    Currently, the implementation of this class does not add anything to
    `~brian2.equations.codestrings.CodeString`, but it should be used instead
    of that class for clarity and to allow for future functionality that is
    only relevant to statements and not to expressions.
    """

    pass


class Expression(CodeString):
    """
    Class for representing an expression.

    Parameters
    ----------
    code : str, optional
        The expression. Note that the expression has to be written in a form
        that is parseable by sympy. Alternatively, a sympy expression can be
        provided (in the ``sympy_expression`` argument).
    sympy_expression : sympy expression, optional
        A sympy expression. Alternatively, a plain string expression can be
        provided (in the ``code`` argument).
    """

    def __init__(self, code=None, sympy_expression=None):
        if code is None and sympy_expression is None:
            raise TypeError("Have to provide either a string or a sympy expression")
        if code is not None and sympy_expression is not None:
            raise TypeError(
                "Provide a string expression or a sympy expression, not both"
            )

        if code is None:
            code = sympy_to_str(sympy_expression)
        else:
            # Just try to convert it to a sympy expression to get syntax errors
            # for incorrect expressions
            str_to_sympy(code)
        super().__init__(code=code)

    stochastic_variables = property(
        lambda self: {
            variable
            for variable in self.identifiers
            if variable == "xi" or variable.startswith("xi_")
        },
        doc="Stochastic variables in this expression",
    )

    def split_stochastic(self):
        """
        Split the expression into a stochastic and non-stochastic part.

        Splits the expression into a tuple of one `Expression` objects f (the
        non-stochastic part) and a dictionary mapping stochastic variables
        to `Expression` objects. For example, an expression of the form
        ``f + g * xi_1 + h * xi_2`` would be returned as:
        ``(f, {'xi_1': g, 'xi_2': h})``
        Note that the `Expression` objects for the stochastic parts do not
        include the stochastic variable itself.

        Returns
        -------
        (f, d) : (`Expression`, dict)
            A tuple of an `Expression` object and a dictionary, the first
            expression being the non-stochastic part of the equation and
            the dictionary mapping stochastic variables (``xi`` or starting
            with ``xi_``) to `Expression` objects. If no stochastic variable
            is present in the code string, a tuple ``(self, None)`` will be
            returned with the unchanged `Expression` object.
        """
        stochastic_variables = []
        for identifier in self.identifiers:
            if identifier == "xi" or identifier.startswith("xi_"):
                stochastic_variables.append(identifier)

        # No stochastic variable
        if not len(stochastic_variables):
            return (self, None)

        stochastic_symbols = [
            sympy.Symbol(variable, real=True) for variable in stochastic_variables
        ]

        # Note that collect only works properly if the expression is expanded
        collected = (
            str_to_sympy(self.code).expand().collect(stochastic_symbols, evaluate=False)
        )

        f_expr = None
        stochastic_expressions = {}
        for var, s_expr in collected.items():
            expr = Expression(sympy_expression=s_expr)
            if var == 1:
                if any(s_expr.has(s) for s in stochastic_symbols):
                    raise AssertionError(
                        "Error when separating expression "
                        f"'{self.code}' into stochastic and non-"
                        "stochastic term: non-stochastic "
                        f"part was determined to be '{s_expr}' but "
                        "contains a stochastic symbol."
                    )
                f_expr = expr
            elif var in stochastic_symbols:
                stochastic_expressions[str(var)] = expr
            else:
                raise ValueError(
                    f"Expression '{self.code}' cannot be separated into "
                    "stochastic and non-stochastic "
                    "term"
                )

        if f_expr is None:
            f_expr = Expression("0.0")

        return f_expr, stochastic_expressions

    def _repr_pretty_(self, p, cycle):
        """
        Pretty printing for ipython.
        """
        if cycle:
            raise AssertionError("Cyclical call of 'CodeString._repr_pretty'")
        # Make use of sympy's pretty printing
        p.pretty(str_to_sympy(self.code))

    def __eq__(self, other):
        if not isinstance(other, Expression):
            return NotImplemented
        return self.code == other.code

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.code)


def is_constant_over_dt(expression, variables, dt_value):
    """
    Check whether an expression can be considered as constant over a time step.
    This is *not* the case when the expression either:

    1. contains the variable ``t`` (except as the argument of a function that
       can be considered as constant over a time step, e.g. a `TimedArray` with
       a dt equal to or greater than the dt used to evaluate this expression)
    2. refers to a stateful function such as ``rand()``.

    Parameters
    ----------
    expression : `sympy.Expr`
        The (sympy) expression to analyze
    variables : dict
        The variables dictionary.
    dt_value : float or None
        The length of a timestep (without units), can be ``None`` if the
        time step is not yet known.

    Returns
    -------
    is_constant : bool
        Whether the expression can be considered to be constant over a time
        step.
    """
    t_symbol = sympy.Symbol("t", real=True, positive=True)
    if expression == t_symbol:
        return False  # The full expression is simply "t"
    func_name = str(expression.func)
    func_variable = variables.get(func_name, None)
    if func_variable is not None and not func_variable.stateless:
        return False
    for arg in expression.args:
        if arg == t_symbol and dt_value is not None:
            # We found "t" -- if it is not the only argument of a locally
            # constant function we bail out
            if not (
                func_variable is not None
                and func_variable.is_locally_constant(dt_value)
            ):
                return False
        else:
            if not is_constant_over_dt(arg, variables, dt_value):
                return False
    return True



#!/usr/bin/env python
"""
Wang-Buszaki model
------------------

J Neurosci. 1996 Oct 15;16(20):6402-13.
Gamma oscillation by synaptic inhibition in a hippocampal interneuronal network model.
Wang XJ, Buzsaki G.

Note that implicit integration (exponential Euler) cannot be used, and therefore
simulation is rather slow.
"""
from brian2 import *

defaultclock.dt = 0.01*ms

Cm = 1*uF # /cm**2
Iapp = 2*uA
gL = 0.1*msiemens
EL = -65*mV
ENa = 55*mV
EK = -90*mV
gNa = 35*msiemens
gK = 9*msiemens

eqs = '''
dv/dt = (-gNa*m**3*h*(v-ENa)-gK*n**4*(v-EK)-gL*(v-EL)+Iapp)/Cm : volt
m = alpha_m/(alpha_m+beta_m) : 1
alpha_m = 0.1/mV*10*mV/exprel(-(v+35*mV)/(10*mV))/ms : Hz
beta_m = 4*exp(-(v+60*mV)/(18*mV))/ms : Hz
dh/dt = 5*(alpha_h*(1-h)-beta_h*h) : 1
alpha_h = 0.07*exp(-(v+58*mV)/(20*mV))/ms : Hz
beta_h = 1./(exp(-0.1/mV*(v+28*mV))+1)/ms : Hz
dn/dt = 5*(alpha_n*(1-n)-beta_n*n) : 1
alpha_n = 0.01/mV*10*mV/exprel(-(v+34*mV)/(10*mV))/ms : Hz
beta_n = 0.125*exp(-(v+44*mV)/(80*mV))/ms : Hz
'''

neuron = NeuronGroup(1, eqs, method='exponential_euler')
neuron.v = -70*mV
neuron.h = 1
M = StateMonitor(neuron, 'v', record=0)

run(100*ms, report='text')

plot(M.t/ms, M[0].v/mV)
show()



'''
How long does synapse creation take?
'''
import time
import pickle

import numpy as np
import joblib

from brian2 import *

repetitions = 3

memory = joblib.Memory(cachedir='.', verbose=0)

@memory.cache
def test_connectivity2(N, i, j, n, p, codeobj_class):
    G = NeuronGroup(N, '')
    # Do it once without measuring the time to ignore the compilation time for
    # C code
    S = Synapses(G, G, '', codeobj_class=codeobj_class)
    S.connect(i, j, p, n)
    connections = len(S)
    del S
    times = []
    for _ in range(repetitions):
        S = Synapses(G, G, '', codeobj_class=codeobj_class)
        start = time.time()
        S.connect(i, j, p, n)
        times.append(time.time() - start)
        del S

    return float(np.median(times)), connections

conditions = [('Full', 'True'),
              ('Full (no-self)', 'i != j'),
              ('One-to-one', 'i == j'),
              ('Simple neighbourhood', 'abs(i-j) < 5'),
              ('Gauss neighbourhood', 'exp(-(i - j)**2/5) > 0.005'),
              ('Random (50%)', (True, None, 1, 0.5)),
              ('Random (10%)', (True, None, 1, 0.1)),
              ('Random (1%)', (True, None, 1, 0.01)),
              ('Random no-self (50%)', ('(i != j)', None, 1, 0.5)),
              ('Random no-self (10%)', ('(i != j)', None, 1, 0.1)),
              ('Random no-self (1%)', ('(i != j)', None, 1, 0.01))]
targets = [NumpyCodeObject]
results = {}
max_connections = 2500000
for target in targets:
    lang_name = target.class_name
    for pattern, condition in conditions:
        N = 1
        connections = took = 0
        while connections < max_connections and took < 20:
            print(lang_name, pattern)
            if isinstance(condition, str):
                took, connections = test_connectivity2(N, condition, None, 1, 1.,
                                                      codeobj_class=target)
            else:
                took, connections = test_connectivity2(N, *condition,
                                                      codeobj_class=target)
            print(N, '%.4fs (for %d connections)' % (took, connections))
            results[(lang_name, connections, pattern)] = took
            N *= 2

with open('synapse_creation_times_brian2.pickle', 'w') as f:
    pickle.dump(results, f)



"""
A simple example of using `Synapses`.
"""
from brian2 import *

G1 = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1',
                 threshold='v > 1', reset='v=0.', method='exact')
G1.v = 1.2
G2 = NeuronGroup(10, 'dv/dt = -v / (10*ms) : 1',
                 threshold='v > 1', reset='v=0', method='exact')
 
syn = Synapses(G1, G2, 'dw/dt = -w / (50*ms): 1 (event-driven)', on_pre='v += w')

syn.connect('i == j', p=0.75)

# Set the delays
syn.delay = '1*ms + i*ms + 0.25*ms * randn()'
# Set the initial values of the synaptic variable
syn.w = 1

mon = StateMonitor(G2, 'v', record=True)
run(20*ms)
plot(mon.t/ms, mon.v.T)
xlabel('Time (ms)')
ylabel('v')
show()



"""
Implementation of `PoissonInput`.
"""

from brian2.core.variables import Variables
from brian2.groups.group import CodeRunner
from brian2.units.fundamentalunits import (
    DimensionMismatchError,
    check_units,
    get_dimensions,
    have_same_dimensions,
)
from brian2.units.stdunits import Hz

from .binomial import BinomialFunction

__all__ = ["PoissonInput"]


class PoissonInput(CodeRunner):
    """
    PoissonInput(target, target_var, N, rate, weight, when='synapses', order=0)

    Adds independent Poisson input to a target variable of a `Group`. For large
    numbers of inputs, this is much more efficient than creating a
    `PoissonGroup`. The synaptic events are generated randomly during the
    simulation and are not preloaded and stored in memory. All the inputs must
    target the same variable, have the same frequency and same synaptic weight.
    All neurons in the target `Group` receive independent realizations of
    Poisson spike trains.

    Parameters
    ----------
    target : `Group`
        The group that is targeted by this input.
    target_var : str
        The variable of `target` that is targeted by this input.
    N : int
        The number of inputs
    rate : `Quantity`
        The rate of each of the inputs
    weight : str or `Quantity`
        Either a string expression (that can be interpreted in the context of
        `target`) or a `Quantity` that will be added for every event to
        the `target_var` of `target`. The unit has to match the unit of
        `target_var`
    when : str, optional
        When to update the target variable during a time step. Defaults to
        the `synapses` scheduling slot. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of the update compared to other operations occurring at
        the same time step and in the same scheduling slot. Defaults to 0.

    """

    @check_units(N=1, rate=Hz)
    def __init__(self, target, target_var, N, rate, weight, when="synapses", order=0):
        if target_var not in target.variables:
            raise KeyError(f"{target_var} is not a variable of {target.name}")

        self._weight = weight
        self._target_var = target_var

        if isinstance(weight, str):
            weight = f"({weight})"
        else:
            weight_dims = get_dimensions(weight)
            target_dims = target.variables[target_var].dim
            # This will be checked automatically in the abstract code as well
            # but doing an explicit check here allows for a clearer error
            # message
            if not have_same_dimensions(weight_dims, target_dims):
                raise DimensionMismatchError(
                    "The provided weight does not "
                    "have the same unit as the "
                    f"target variable '{target_var}'",
                    weight_dims,
                    target_dims,
                )
            weight = repr(weight)
        self._N = N
        self._rate = rate
        binomial_sampling = BinomialFunction(
            N, rate * target.clock.dt, name="poissoninput_binomial*"
        )

        code = f"{target_var} += {binomial_sampling.name}()*{weight}"
        self._stored_dt = target.dt_[:]  # make a copy
        # FIXME: we need an explicit reference here for on-the-fly subgroups
        # For example: PoissonInput(group[:N], ...)
        self._group = target
        CodeRunner.__init__(
            self,
            group=target,
            template="stateupdate",
            code=code,
            user_code="",
            when=when,
            order=order,
            name="poissoninput*",
            clock=target.clock,
        )
        self.variables = Variables(self)
        self.variables._add_variable(binomial_sampling.name, binomial_sampling)

    rate = property(fget=lambda self: self._rate, doc="The rate of each input")
    N = property(fget=lambda self: self._N, doc="The number of inputs")
    target_var = property(
        fget=lambda self: self._target_var, doc="The targetted variable"
    )
    weight = property(fget=lambda self: self._weight, doc="The synaptic weight")

    def before_run(self, run_namespace):
        if self._group.dt_ != self._stored_dt:
            raise NotImplementedError(
                f"The dt used for simulating {self.group.name} "
                "changed after the PoissonInput source was "
                "created."
            )
        CodeRunner.before_run(self, run_namespace=run_namespace)



#!/usr/bin/env python
"""
Chaos in the AdEx model
-----------------------
Fig. 8B from:
Touboul, J. and Brette, R. (2008). Dynamics and bifurcations of the adaptive
exponential integrate-and-fire model. Biological Cybernetics 99(4-5):319-34.

This shows the bifurcation structure when the reset value is varied
(vertical axis shows the values of w at spike times for a given a reset value
Vr).
"""
from brian2 import *

defaultclock.dt = 0.01*ms

C = 281*pF
gL = 30*nS
EL = -70.6*mV
VT = -50.4*mV
DeltaT = 2*mV
tauw = 40*ms
a = 4*nS
b = 0.08*nA
I = .8*nA
Vcut = VT + 5 * DeltaT  # practical threshold condition
N = 200

eqs = """
dvm/dt=(gL*(EL-vm)+gL*DeltaT*exp((vm-VT)/DeltaT)+I-w)/C : volt
dw/dt=(a*(vm-EL)-w)/tauw : amp
Vr:volt
"""

neuron = NeuronGroup(N, model=eqs, threshold='vm > Vcut',
                     reset="vm = Vr; w += b", method='euler')
neuron.vm = EL
neuron.w = a * (neuron.vm - EL)
neuron.Vr = linspace(-48.3 * mV, -47.7 * mV, N)  # bifurcation parameter

init_time = 3*second
run(init_time, report='text')  # we discard the first spikes

states = StateMonitor(neuron, "w", record=True, when='start')
spikes = SpikeMonitor(neuron)
run(1 * second, report='text')

# Get the values of Vr and w for each spike
Vr = neuron.Vr[spikes.i]
w = states.w[spikes.i, int_((spikes.t-init_time)/defaultclock.dt)]

figure()
plot(Vr / mV, w / nA, '.k')
xlabel('Vr (mV)')
ylabel('w (nA)')
show()



"""
Set state variable values with a string (using code generation).
"""

from brian2 import *

G = NeuronGroup(100, 'v:volt', threshold='v>-50*mV')
G.v = '(sin(2*pi*i/N) - 70 + 0.25*randn()) * mV'
S = Synapses(G, G, 'w : volt', on_pre='v += w')
S.connect()

space_constant = 200.0
S.w['i > j'] = 'exp(-(i - j)**2/space_constant) * mV'

# Generate a matrix for display
w_matrix = np.zeros((len(G), len(G)))
w_matrix[S.i[:], S.j[:]] = S.w[:]

subplot(1, 2, 1)
plot(G.v[:] / mV)
xlabel('Neuron index')
ylabel('v')
subplot(1, 2, 2)
imshow(w_matrix)
xlabel('i')
ylabel('j')
title('Synaptic weight')
tight_layout()
show()



"""
Module defining `PopulationRateMonitor`.
"""

import numpy as np

from brian2.core.variables import Variables
from brian2.groups.group import CodeRunner, Group
from brian2.units.allunits import hertz, second
from brian2.units.fundamentalunits import Quantity, check_units
from brian2.utils.logger import get_logger

__all__ = ["PopulationRateMonitor"]


logger = get_logger(__name__)


class PopulationRateMonitor(Group, CodeRunner):
    """
    Record instantaneous firing rates, averaged across neurons from a
    `NeuronGroup` or other spike source.

    Parameters
    ----------
    source : (`NeuronGroup`, `SpikeSource`)
        The source of spikes to record.
    name : str, optional
        A unique name for the object, otherwise will use
        ``source.name+'_ratemonitor_0'``, etc.
    codeobj_class : class, optional
        The `CodeObject` class to run code with.
    dtype : dtype, optional
        The dtype to use to store the ``rate`` variable. Defaults to
        `~numpy.float64`, i.e. double precision.
    Notes
    -----
    Currently, this monitor can only monitor the instantaneous firing rates at
    each time step of the source clock. Any binning/smoothing of the firing
    rates has to be done manually afterwards.
    """

    invalidates_magic_network = False
    add_to_magic_network = True

    def __init__(
        self, source, name="ratemonitor*", codeobj_class=None, dtype=np.float64
    ):
        #: The group we are recording from
        self.source = source

        self.codeobj_class = codeobj_class
        CodeRunner.__init__(
            self,
            group=self,
            code="",
            template="ratemonitor",
            clock=source.clock,
            when="end",
            order=0,
            name=name,
        )

        self.add_dependency(source)

        self.variables = Variables(self)
        # Handle subgroups correctly
        start = getattr(source, "start", 0)
        stop = getattr(source, "stop", len(source))
        self.variables.add_constant("_source_start", start)
        self.variables.add_constant("_source_stop", stop)
        self.variables.add_reference("_spikespace", source)
        self.variables.add_dynamic_array(
            "rate", size=0, dimensions=hertz.dim, read_only=True, dtype=dtype
        )
        self.variables.add_dynamic_array(
            "t",
            size=0,
            dimensions=second.dim,
            read_only=True,
            dtype=self._clock.variables["t"].dtype,
        )
        self.variables.add_reference("_num_source_neurons", source, "N")
        self.variables.add_array(
            "N", dtype=np.int32, size=1, scalar=True, read_only=True
        )
        self.variables.create_clock_variables(self._clock, prefix="_clock_")
        self._enable_group_attributes()

    def resize(self, new_size):
        # Note that this does not set N, this has to be done in the template
        # since we use a restricted pointer to access it (which promises that
        # we only change the value through this pointer)
        self.variables["rate"].resize(new_size)
        self.variables["t"].resize(new_size)

    def reinit(self):
        """
        Clears all recorded rates
        """
        raise NotImplementedError()

    @check_units(width=second)
    def smooth_rate(self, window="gaussian", width=None):
        """
        smooth_rate(self, window='gaussian', width=None)

        Return a smooth version of the population rate.

        Parameters
        ----------
        window : str, ndarray
            The window to use for smoothing. Can be a string to chose a
            predefined window(``'flat'`` for a rectangular, and ``'gaussian'``
            for a Gaussian-shaped window). In this case the width of the window
            is determined by the ``width`` argument. Note that for the Gaussian
            window, the ``width`` parameter specifies the standard deviation of
            the Gaussian, the width of the actual window is ``4*width + dt``
            (rounded to the nearest dt). For the flat window, the width is
            rounded to the nearest odd multiple of dt to avoid shifting the rate
            in time.
            Alternatively, an arbitrary window can be given as a numpy array
            (with an odd number of elements). In this case, the width in units
            of time depends on the ``dt`` of the simulation, and no ``width``
            argument can be specified. The given window will be automatically
            normalized to a sum of 1.
        width : `Quantity`, optional
            The width of the ``window`` in seconds (for a predefined window).

        Returns
        -------
        rate : `Quantity`
            The population rate in Hz, smoothed with the given window. Note that
            the rates are smoothed and not re-binned, i.e. the length of the
            returned array is the same as the length of the ``rate`` attribute
            and can be plotted against the `PopulationRateMonitor` 's ``t``
            attribute.
        """
        if width is None and isinstance(window, str):
            raise TypeError("Need a width when using a predefined window.")
        if width is not None and not isinstance(window, str):
            raise TypeError("Can only specify a width for a predefined window")

        if isinstance(window, str):
            if window == "gaussian":
                width_dt = int(np.round(2 * width / self.clock.dt))
                # Rounding only for the size of the window, not for the standard
                # deviation of the Gaussian
                window = np.exp(
                    -np.arange(-width_dt, width_dt + 1) ** 2
                    * 1.0
                    / (2 * (width / self.clock.dt) ** 2)
                )
            elif window == "flat":
                width_dt = int(width / 2 / self.clock.dt) * 2 + 1
                used_width = width_dt * self.clock.dt
                if abs(used_width - width) > 1e-6 * self.clock.dt:
                    logger.info(
                        f"width adjusted from {width} to {used_width}",
                        "adjusted_width",
                        once=True,
                    )
                window = np.ones(width_dt)
            else:
                raise NotImplementedError(f'Unknown pre-defined window "{window}"')
        else:
            try:
                window = np.asarray(window)
            except TypeError:
                raise TypeError(f"Cannot use a window of type {type(window)}")
            if window.ndim != 1:
                raise TypeError("The provided window has to be one-dimensional.")
            if len(window) % 2 != 1:
                raise TypeError("The window has to have an odd number of values.")
        return Quantity(
            np.convolve(self.rate_, window * 1.0 / sum(window), mode="same"),
            dim=hertz.dim,
        )

    def __repr__(self):
        classname = self.__class__.__name__
        return f"<{classname}, recording {self.source.name}>"



"""
Module defining `SpikeGeneratorGroup`.
"""

import numpy as np

from brian2.core.functions import timestep
from brian2.core.spikesource import SpikeSource
from brian2.core.variables import Variables
from brian2.groups.group import CodeRunner, Group
from brian2.units.allunits import second
from brian2.units.fundamentalunits import Quantity, check_units
from brian2.utils.logger import get_logger

__all__ = ["SpikeGeneratorGroup"]


logger = get_logger(__name__)


class SpikeGeneratorGroup(Group, CodeRunner, SpikeSource):
    """
    SpikeGeneratorGroup(N, indices, times, dt=None, clock=None,
                        period=0*second, when='thresholds', order=0,
                        sorted=False, name='spikegeneratorgroup*',
                        codeobj_class=None)

    A group emitting spikes at given times.

    Parameters
    ----------
    N : int
        The number of "neurons" in this group
    indices : array of integers
        The indices of the spiking cells
    times : `Quantity`
        The spike times for the cells given in ``indices``. Has to have the
        same length as ``indices``.
    period : `Quantity`, optional
        If this is specified, it will repeat spikes with this period. A
        period of 0s means not repeating spikes.
    dt : `Quantity`, optional
        The time step to be used for the simulation. Cannot be combined with
        the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    when : str, optional
        When to run within a time step, defaults to the ``'thresholds'`` slot.
        See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of of this group for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    sorted : bool, optional
        Whether the given indices and times are already sorted. Set to ``True``
        if your events are already sorted (first by spike time, then by index),
        this can save significant time at construction if your arrays contain
        large numbers of spikes. Defaults to ``False``.

    Notes
    -----
    * If `sorted` is set to ``True``, the given arrays will not be copied
      (only affects runtime mode)..
    """

    @check_units(N=1, indices=1, times=second, period=second)
    def __init__(
        self,
        N,
        indices,
        times,
        dt=None,
        clock=None,
        period=0 * second,
        when="thresholds",
        order=0,
        sorted=False,
        name="spikegeneratorgroup*",
        codeobj_class=None,
    ):
        Group.__init__(self, dt=dt, clock=clock, when=when, order=order, name=name)

        # We store the indices and times also directly in the Python object,
        # this way we can use them for checks in `before_run` even in standalone
        # TODO: Remove this when the checks in `before_run` have been moved to the template
        #: Array of spiking neuron indices.
        self._neuron_index = None
        #: Array of spiking neuron times.
        self._spike_time = None
        #: "Dirty flag" that will be set when spikes are changed after the
        #: `before_run` check
        self._spikes_changed = True

        # Let other objects know that we emit spikes events
        self.events = {"spike": None}

        self.codeobj_class = codeobj_class

        if N < 1 or int(N) != N:
            raise TypeError("N has to be an integer >=1.")
        N = int(N)
        self.start = 0
        self.stop = N

        self.variables = Variables(self)
        self.variables.create_clock_variables(self._clock)

        indices, times = self._check_args(
            indices, times, period, N, sorted, self._clock.dt
        )

        self.variables.add_constant("N", value=N)
        self.variables.add_array(
            "period",
            dimensions=second.dim,
            size=1,
            constant=True,
            read_only=True,
            scalar=True,
            dtype=self._clock.variables["t"].dtype,
        )
        self.variables.add_arange("i", N)
        self.variables.add_dynamic_array(
            "spike_number",
            values=np.arange(len(indices)),
            size=len(indices),
            dtype=np.int32,
            read_only=True,
            constant=True,
            index="spike_number",
            unique=True,
        )
        self.variables.add_dynamic_array(
            "neuron_index",
            values=indices,
            size=len(indices),
            dtype=np.int32,
            index="spike_number",
            read_only=True,
            constant=True,
        )
        self.variables.add_dynamic_array(
            "spike_time",
            values=times,
            size=len(times),
            dimensions=second.dim,
            index="spike_number",
            read_only=True,
            constant=True,
            dtype=self._clock.variables["t"].dtype,
        )
        self.variables.add_dynamic_array(
            "_timebins",
            size=len(times),
            index="spike_number",
            read_only=True,
            constant=True,
            dtype=np.int32,
        )
        self.variables.add_array(
            "_period_bins",
            size=1,
            constant=True,
            read_only=True,
            scalar=True,
            dtype=np.int32,
        )
        self.variables.add_array("_spikespace", size=N + 1, dtype=np.int32)
        self.variables.add_array(
            "_lastindex", size=1, values=0, dtype=np.int32, read_only=True, scalar=True
        )

        #: Remember the dt we used the last time when we checked the spike bins
        #: to not repeat the work for multiple runs with the same dt
        self._previous_dt = None

        CodeRunner.__init__(
            self,
            self,
            code="",
            template="spikegenerator",
            clock=self._clock,
            when=when,
            order=order,
            name=None,
        )

        # Activate name attribute access
        self._enable_group_attributes()

        self.variables["period"].set_value(period)

    def _full_state(self):
        state = super()._full_state()
        # Store the internal information we use to decide whether to rebuild
        # the time bins
        state["_previous_dt"] = self._previous_dt
        state["_spikes_changed"] = self._spikes_changed
        return state

    def _restore_from_full_state(self, state):
        state = state.copy()  # copy to avoid errors for multiple restores
        self._previous_dt = state.pop("_previous_dt")
        self._spikes_changed = state.pop("_spikes_changed")
        super()._restore_from_full_state(state)

    def before_run(self, run_namespace):
        # Do some checks on the period vs. dt
        dt = self.dt_[:]  # make a copy
        period = self.period_
        if period < np.inf and period != 0:
            if period < dt:
                raise ValueError(
                    f"The period of '{self.name}' is {self.period[:]!s}, "
                    f"which is smaller than its dt of {dt*second!s}."
                )

        if self._spikes_changed:
            current_t = self.variables["t"].get_value().item()
            timesteps = timestep(self._spike_time, dt)
            current_step = timestep(current_t, dt)
            in_the_past = np.nonzero(timesteps < current_step)[0]
            if len(in_the_past):
                logger.warn(
                    "The SpikeGeneratorGroup contains spike times "
                    "earlier than the start time of the current run "
                    f"(t = {current_t*second!s}), these spikes will be "
                    "ignored.",
                    name_suffix="ignored_spikes",
                )
                self.variables["_lastindex"].set_value(in_the_past[-1] + 1)
            else:
                self.variables["_lastindex"].set_value(0)

        # Check that we don't have more than one spike per neuron in a time bin
        if self._previous_dt is None or dt != self._previous_dt or self._spikes_changed:
            # We shift all the spikes by a tiny amount to make sure that spikes
            # at exact multiples of dt do not end up in the previous time bin
            # This shift has to be quite significant relative to machine
            # epsilon, we use 1e-3 of the dt here
            shift = 1e-3 * dt
            timebins = np.asarray(
                np.asarray(self._spike_time + shift) / dt, dtype=np.int32
            )
            # time is already in sorted order, so it's enough to check if the condition
            # that timebins[i]==timebins[i+1] and self._neuron_index[i]==self._neuron_index[i+1]
            # is ever both true
            if (
                np.logical_and(np.diff(timebins) == 0, np.diff(self._neuron_index) == 0)
            ).any():
                raise ValueError(
                    f"Using a dt of {self.dt!s}, some neurons of "
                    f"SpikeGeneratorGroup '{self.name}' spike more than "
                    "once during a time step."
                )
            self.variables["_timebins"].set_value(timebins)
            period_bins = np.round(period / dt)
            max_int = np.iinfo(np.int32).max
            if period_bins > max_int:
                logger.warn(
                    f"Periods longer than {max_int} timesteps "
                    f"(={max_int*dt*second!s}) are not "
                    "supported, the period will therefore be "
                    "considered infinite. Set the period to 0*second "
                    "to avoid this "
                    "warning.",
                    "spikegenerator_long_period",
                )
                period = period_bins = 0
            if np.abs(period_bins * dt - period) > period * np.finfo(dt.dtype).eps:
                raise NotImplementedError(
                    f"The period of '{self.name}' is "
                    f"{self.period[:]!s}, which is "
                    "not an integer multiple of its dt "
                    f"of {dt*second!s}."
                )

            self.variables["_period_bins"].set_value(period_bins)

            self._previous_dt = dt
            self._spikes_changed = False

        super().before_run(run_namespace=run_namespace)

    @check_units(indices=1, times=second, period=second)
    def set_spikes(self, indices, times, period=0 * second, sorted=False):
        """
        set_spikes(indices, times, period=0*second, sorted=False)

        Change the spikes that this group will generate.

        This can be used to set the input for a second run of a model based on
        the output of a first run (if the input for the second run is already
        known before the first run, then all the information should simply be
        included in the initial `SpikeGeneratorGroup` initializer call,
        instead).

        Parameters
        ----------
        indices : array of integers
            The indices of the spiking cells
        times : `Quantity`
            The spike times for the cells given in ``indices``. Has to have the
            same length as ``indices``.
        period : `Quantity`, optional
            If this is specified, it will repeat spikes with this period. A
            period of 0s means not repeating spikes.
        sorted : bool, optional
            Whether the given indices and times are already sorted. Set to
            ``True`` if your events are already sorted (first by spike time,
            then by index), this can save significant time at construction if
            your arrays contain large numbers of spikes. Defaults to ``False``.
        """

        indices, times = self._check_args(
            indices, times, period, self.N, sorted, self.dt
        )

        self.variables["period"].set_value(period)
        self.variables["neuron_index"].resize(len(indices))
        self.variables["spike_time"].resize(len(indices))
        self.variables["spike_number"].resize(len(indices))
        self.variables["spike_number"].set_value(np.arange(len(indices)))
        self.variables["_timebins"].resize(len(indices))
        self.variables["neuron_index"].set_value(indices)
        self.variables["spike_time"].set_value(times)
        # _lastindex and _timebins will be set as part of before_run

    def _check_args(self, indices, times, period, N, sorted, dt):
        times = Quantity(times)
        if len(indices) != len(times):
            raise ValueError(
                "Length of the indices and times array must "
                f"match, but {len(indices)} != {len(times)}"
            )
        if period < 0 * second:
            raise ValueError("The period cannot be negative.")
        elif len(times) and period != 0 * second:
            period_bins = np.round(period / dt)
            # Note: we have to use the timestep function here, to use the same
            # binning as in the actual simulation
            max_bin = timestep(np.max(times), dt)
            if max_bin >= period_bins:
                raise ValueError(
                    "The period has to be greater than the maximum of the spike times"
                )
        if len(times) and np.min(times) < 0 * second:
            raise ValueError("Spike times cannot be negative")
        if len(indices) and (np.min(indices) < 0 or np.max(indices) >= N):
            raise ValueError(f"Indices have to lie in the interval [0, {int(N)}[")

        times = np.asarray(times)
        indices = np.asarray(indices)
        if not sorted:
            # sort times and indices first by time, then by indices
            sort_indices = np.lexsort((indices, times))
            indices = indices[sort_indices]
            times = times[sort_indices]

        # We store the indices and times also directly in the Python object,
        # this way we can use them for checks in `before_run` even in standalone
        # TODO: Remove this when the checks in `before_run` have been moved to the template
        self._neuron_index = indices
        self._spike_time = times
        self._spikes_changed = True

        return indices, times

    @property
    def spikes(self):
        """
        The spikes returned by the most recent thresholding operation.
        """
        # Note that we have to directly access the ArrayVariable object here
        # instead of using the Group mechanism by accessing self._spikespace
        # Using the latter would cut _spikespace to the length of the group
        spikespace = self.variables["_spikespace"].get_value()
        return spikespace[: spikespace[-1]]

    def __repr__(self):
        cls = self.__class__.__name__
        size = self.variables["neuron_index"].size
        return (
            f"{cls}({self.N}, indices=<length {size} array>, times=<length"
            f" {size} array>)"
        )



"""
Tests of input features
"""

from brian2 import *
from brian2.tests.features import FeatureTest, InaccuracyError


class SpikeGeneratorGroupTest(FeatureTest):
    category = "Input"
    name = "SpikeGeneratorGroup"
    tags = ["SpikeMonitor", "run", "SpikeGeneratorGroup"]

    def run(self):
        N = 10
        numspikes = 1000
        i = arange(numspikes) % N
        t = linspace(0, 1, numspikes) * (100 * ms)
        G = SpikeGeneratorGroup(N, i, t)
        self.M = M = SpikeMonitor(G)
        run(100 * ms)

    def results(self):
        return {"i": self.M.i[:], "t": self.M.t[:]}

    compare = FeatureTest.compare_arrays



{# USES_VARIABLES { N } #}
{% extends 'common_group.pyx' %}

{% block maincode %}

    {%set c_type = cpp_dtype(variables['_variable'].dtype) %}
    {%set np_type = numpy_dtype(variables['_variable'].dtype) %}

    _vectorisation_idx = 1

    cdef size_t _N = {{constant_or_scalar('N', variables['N'])}}
    cdef size_t _num_elements = 0
    cdef _numpy.ndarray[{{c_type}}, ndim=1, mode='c'] _elements = _numpy.zeros(_N, dtype=_numpy.{{np_type}})
    cdef {{c_type}}[:] _elements_view = _elements


    {{scalar_code|autoindent}}

    for _idx in range(_N):
        _vectorisation_idx = _idx
        
        {{vector_code|autoindent}}
        
        if _cond:
            _elements_view[_num_elements] = _variable
            _num_elements += 1
    
    return _elements[:_num_elements]

{% endblock %}



#!/usr/bin/env python3
"""Brian2 simulation of spike-based homeostasis

Following
O. Breitwieser: "Towards a Neuromorphic Implementation of Spike-Based Expectation Maximization"

Two poisson stimuli are connected to a neuron. One with a varying rate
and the other with a fixed rate.  The synaptic weight from the varying
rate stimulus to the neuron is fixed. The synaptic weight from the
fixed rate stimulus to the neuron is plastic and tries to keep the
neuron at a firing rate that is determined by the parameters of the
plasticity rule.

Sebastian Schmitt, 2021
"""

import itertools
import numpy as np
import matplotlib.pyplot as plt

from brian2 import TimedArray, PoissonGroup, NeuronGroup, Synapses, StateMonitor, PopulationRateMonitor
from brian2 import defaultclock, run
from brian2 import Hz, ms, second

# The synaptic weight from the steady stimulus is plastic
steady_stimulus = TimedArray([50]*Hz, dt=40*second)
steady_poisson = PoissonGroup(1, rates='steady_stimulus(t)')

# The synaptic weight from the varying stimulus is static
varying_stimulus = TimedArray([25*Hz, 50*Hz, 0*Hz, 35*Hz, 0*Hz], dt=10*second)
varying_poisson = PoissonGroup(1, rates='varying_stimulus(t)')

# dw_plus/dw_minus determines scales the steady stimulus rate to the target firing rate, must not be larger 1
# the magntude of dw_plus and dw_minus determines the "speed" of the homeostasis
parameters = {
    'tau': 10*ms,  # membrane time constant
    'dw_plus': 0.05,  # weight increment on pre spike
    'dw_minus': 0.05,  # weight increment on post spike
    'w_max': 2,  # maximum plastic weight
    'w_initial': 0  # initial plastic weight
}

eqs = 'dv/dt = (0 - v)/tau : 1 (unless refractory)'

neuron_with_homeostasis = NeuronGroup(1, eqs,
                                      threshold='v > 1', reset='v = -1',
                                      method='euler', refractory=1*ms,
                                      namespace=parameters)
neuron_without_homeostasis = NeuronGroup(1, eqs,
                                         threshold='v > 1', reset='v = -1',
                                         method='euler', refractory=1*ms,
                                         namespace=parameters)

plastic_synapse = Synapses(steady_poisson, neuron_with_homeostasis,
                           'w : 1',
                           on_pre='''
                           v_post += w
                           w = clip(w + dw_plus, 0, w_max)
                           ''',
                           on_post='''
                           w = clip(w - dw_minus, 0, w_max)
                           ''', namespace=parameters)
plastic_synapse.connect()
plastic_synapse.w = parameters['w_initial']

non_plastic_synapse_neuron_without_homeostasis = Synapses(varying_poisson,
                                                          neuron_without_homeostasis,
                                                          'w : 1', on_pre='v_post += w')
non_plastic_synapse_neuron_without_homeostasis.connect()
non_plastic_synapse_neuron_without_homeostasis.w = 2

non_plastic_synapse_neuron = Synapses(varying_poisson, neuron_with_homeostasis,
                                      'w : 1', on_pre='v_post += w')
non_plastic_synapse_neuron.connect()
non_plastic_synapse_neuron.w = 2

M = StateMonitor(neuron_with_homeostasis, 'v', record=True)
M2 = StateMonitor(plastic_synapse, 'w', record=True)
M_rate_neuron_with_homeostasis = PopulationRateMonitor(neuron_with_homeostasis)
M_rate_neuron_without_homeostasis = PopulationRateMonitor(neuron_without_homeostasis)

duration = 40*second
defaultclock.dt = 0.1*ms
run(duration, report='text')

fig, axes = plt.subplots(3, sharex=True)

axes[0].plot(M2.t/second, M2.w[0], label="homeostatic weight")
axes[0].set_ylabel("weight")
axes[0].legend()

# dt is in second
dts = np.arange(0., len(varying_stimulus.values)*varying_stimulus.dt, varying_stimulus.dt)
x = list(itertools.chain(*zip(dts, dts)))
y = list(itertools.chain(*zip(varying_stimulus.values/Hz, varying_stimulus.values/Hz)))
axes[1].plot(x, [0] + y[:-1], label="varying stimulus")
axes[1].set_ylabel("rate [Hz]")
axes[1].legend()

# in ms
smooth_width = 100*ms
axes[2].plot(M_rate_neuron_with_homeostasis.t/second,
             M_rate_neuron_with_homeostasis.smooth_rate(width=smooth_width)/Hz,
             label="with homeostasis")
axes[2].plot(M_rate_neuron_without_homeostasis.t/second,
             M_rate_neuron_without_homeostasis.smooth_rate(width=smooth_width)/Hz,
             label="without homeostasis")
axes[2].set_ylabel("firing rate [Hz]")
axes[2].legend()

plt.xlabel('Time (s)')
plt.show()



# Do the equivalent of "from brian2 import *" for all doctests
import pytest

@pytest.fixture(autouse=True)
def add_brian2(doctest_namespace):
    exec('from brian2 import *', doctest_namespace)
    import brian2
    import numpy as np
    # Always use numpy for doctests
    brian2.prefs['codegen.target'] = 'numpy'
    # Print output changed in numpy 1.14, stick with the old format to
    # avoid doctest failures
    try:
        np.set_printoptions(legacy='1.13')
    except TypeError:
        pass  # using a numpy version < 1.14



"""
Hodgkin-Huxley equations (1952).
"""
from brian2 import *

morpho = Cylinder(length=10*cm, diameter=2*238*um, n=1000, type='axon')

El = 10.613*mV
ENa = 115*mV
EK = -12*mV
gl = 0.3*msiemens/cm**2
gNa0 = 120*msiemens/cm**2
gK = 36*msiemens/cm**2

# Typical equations
eqs = '''
# The same equations for the whole neuron, but possibly different parameter values
# distributed transmembrane current
Im = gl * (El-v) + gNa * m**3 * h * (ENa-v) + gK * n**4 * (EK-v) : amp/meter**2
I : amp (point current) # applied current
dm/dt = alpham * (1-m) - betam * m : 1
dn/dt = alphan * (1-n) - betan * n : 1
dh/dt = alphah * (1-h) - betah * h : 1
alpham = (0.1/mV) * 10*mV/exprel((-v+25*mV)/(10*mV))/ms : Hz
betam = 4 * exp(-v/(18*mV))/ms : Hz
alphah = 0.07 * exp(-v/(20*mV))/ms : Hz
betah = 1/(exp((-v+30*mV) / (10*mV)) + 1)/ms : Hz
alphan = (0.01/mV) * 10*mV/exprel((-v+10*mV)/(10*mV))/ms : Hz
betan = 0.125*exp(-v/(80*mV))/ms : Hz
gNa : siemens/meter**2
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=1*uF/cm**2,
                       Ri=35.4*ohm*cm, method="exponential_euler")
neuron.v = 0*mV
neuron.h = 1
neuron.m = 0
neuron.n = .5
neuron.I = 0
neuron.gNa = gNa0
neuron[5*cm:10*cm].gNa = 0*siemens/cm**2
M = StateMonitor(neuron, 'v', record=True)

run(50*ms, report='text')
neuron.I[0] = 1*uA  # current injection at one end
run(3*ms)
neuron.I = 0*amp
run(100*ms, report='text')
for i in range(75, 125, 1):
    plot(cumsum(neuron.length)/cm, i+(1./60)*M.v[:, i*5]/mV, 'k')
yticks([])
ylabel('Time [major] v (mV) [minor]')
xlabel('Position (cm)')
axis('tight')
show()



"""
Handles loading templates from a directory.
"""

import re
from collections.abc import Mapping

from jinja2 import (
    ChoiceLoader,
    Environment,
    PackageLoader,
    StrictUndefined,
    TemplateNotFound,
)

from brian2.utils.stringtools import get_identifiers, indent, strip_empty_lines

__all__ = ["Templater"]

AUTOINDENT_START = "%%START_AUTOINDENT%%"
AUTOINDENT_END = "%%END_AUTOINDENT%%"


def autoindent(code):
    if isinstance(code, list):
        code = "\n".join(code)
    if not code.startswith("\n"):
        code = f"\n{code}"
    if not code.endswith("\n"):
        code = f"{code}\n"
    return AUTOINDENT_START + code + AUTOINDENT_END


def autoindent_postfilter(code):
    lines = code.split("\n")
    outlines = []
    addspaces = 0
    for line in lines:
        if AUTOINDENT_START in line:
            if addspaces > 0:
                raise SyntaxError("Cannot nest autoindents")
            addspaces = line.find(AUTOINDENT_START)
            line = line.replace(AUTOINDENT_START, "")
        if AUTOINDENT_END in line:
            line = line.replace(AUTOINDENT_END, "")
            addspaces = 0
        outlines.append(" " * addspaces + line)
    return "\n".join(outlines)


def variables_to_array_names(variables, access_data=True):
    from brian2.devices.device import get_device

    device = get_device()
    names = [device.get_array_name(var, access_data=access_data) for var in variables]
    return names


class LazyTemplateLoader:
    """
    Helper object to load templates only when they are needed.
    """

    def __init__(self, environment, extension):
        self.env = environment
        self.extension = extension
        self._templates = {}

    def get_template(self, name):
        if name not in self._templates:
            try:
                template = CodeObjectTemplate(
                    self.env.get_template(name + self.extension),
                    self.env.loader.get_source(self.env, name + self.extension)[0],
                )
            except TemplateNotFound:
                try:
                    # Try without extension as well (e.g. for makefiles)
                    template = CodeObjectTemplate(
                        self.env.get_template(name),
                        self.env.loader.get_source(self.env, name)[0],
                    )
                except TemplateNotFound:
                    raise KeyError(f'No template with name "{name}" found.')
            self._templates[name] = template
        return self._templates[name]


class Templater:
    """
    Class to load and return all the templates a `CodeObject` defines.

    Parameters
    ----------

    package_name : str, tuple of str
        The package where the templates are saved. If this is a tuple then each template will be searched in order
        starting from the first package in the tuple until the template is found. This allows for derived templates
        to be used. See also `~Templater.derive`.
    extension : str
        The file extension (e.g. ``.pyx``) used for the templates.
    env_globals : dict (optional)
        A dictionary of global values accessible by the templates. Can be used for providing utility functions.
        In all cases, the filter 'autoindent' is available (see existing templates for example usage).
    templates_dir : str, tuple of str, optional
        The name of the directory containing the templates. Defaults to ``'templates'``.

    Notes
    -----
    Templates are accessed using ``templater.template_base_name`` (the base name is without the file extension).
    This returns a `CodeObjectTemplate`.
    """

    def __init__(
        self, package_name, extension, env_globals=None, templates_dir="templates"
    ):
        if isinstance(package_name, str):
            package_name = (package_name,)
        if isinstance(templates_dir, str):
            templates_dir = (templates_dir,)
        loader = ChoiceLoader(
            [
                PackageLoader(name, t_dir)
                for name, t_dir in zip(package_name, templates_dir)
            ]
        )
        self.env = Environment(
            loader=loader,
            trim_blocks=True,
            lstrip_blocks=True,
            undefined=StrictUndefined,
        )
        self.env.globals["autoindent"] = autoindent
        self.env.filters["autoindent"] = autoindent
        self.env.filters["variables_to_array_names"] = variables_to_array_names
        if env_globals is not None:
            self.env.globals.update(env_globals)
        else:
            env_globals = {}
        self.env_globals = env_globals
        self.package_names = package_name
        self.templates_dir = templates_dir
        self.extension = extension
        self.templates = LazyTemplateLoader(self.env, extension)

    def __getattr__(self, item):
        return self.templates.get_template(item)

    def derive(
        self, package_name, extension=None, env_globals=None, templates_dir="templates"
    ):
        """
        Return a new Templater derived from this one, where the new package name and globals overwrite the old.
        """
        if extension is None:
            extension = self.extension
        if isinstance(package_name, str):
            package_name = (package_name,)
        if env_globals is None:
            env_globals = {}
        if isinstance(templates_dir, str):
            templates_dir = (templates_dir,)
        package_name = package_name + self.package_names
        templates_dir = templates_dir + self.templates_dir
        new_env_globals = self.env_globals.copy()
        new_env_globals.update(**env_globals)
        return Templater(
            package_name,
            extension=extension,
            env_globals=new_env_globals,
            templates_dir=templates_dir,
        )


class CodeObjectTemplate:
    """
    Single template object returned by `Templater` and used for final code generation

    Should not be instantiated by the user, but only directly by `Templater`.

    Notes
    -----

    The final code is obtained from this by calling the template (see `~CodeObjectTemplater.__call__`).
    """

    def __init__(self, template, template_source):
        self.template = template
        self.template_source = template_source
        #: The set of variables in this template
        self.variables = set()
        #: The indices over which the template iterates completely
        self.iterate_all = set()
        #: Read-only variables that are changed by this template
        self.writes_read_only = set()
        # This is the bit inside {} for USES_VARIABLES { list of words }
        specifier_blocks = re.findall(
            r"\bUSES_VARIABLES\b\s*\{(.*?)\}", template_source, re.M | re.S
        )
        # Same for ITERATE_ALL
        iterate_all_blocks = re.findall(
            r"\bITERATE_ALL\b\s*\{(.*?)\}", template_source, re.M | re.S
        )
        # And for WRITES_TO_READ_ONLY_VARIABLES
        writes_read_only_blocks = re.findall(
            r"\bWRITES_TO_READ_ONLY_VARIABLES\b\s*\{(.*?)\}",
            template_source,
            re.M | re.S,
        )
        #: Does this template allow writing to scalar variables?
        self.allows_scalar_write = "ALLOWS_SCALAR_WRITE" in template_source

        for block in specifier_blocks:
            self.variables.update(get_identifiers(block))
        for block in iterate_all_blocks:
            self.iterate_all.update(get_identifiers(block))
        for block in writes_read_only_blocks:
            self.writes_read_only.update(get_identifiers(block))

    def __call__(self, scalar_code, vector_code, **kwds):
        """
        Return a usable code block or blocks from this template.

        Parameters
        ----------
        scalar_code : dict
            Dictionary of scalar code blocks.
        vector_code : dict
            Dictionary of vector code blocks
        **kwds
            Additional parameters to pass to the template

        Notes
        -----

        Returns either a string (if macros were not used in the template), or a `MultiTemplate` (if macros were used).
        """
        if (
            scalar_code is not None
            and len(scalar_code) == 1
            and list(scalar_code)[0] is None
        ):
            scalar_code = scalar_code[None]
        if (
            vector_code is not None
            and len(vector_code) == 1
            and list(vector_code)[0] is None
        ):
            vector_code = vector_code[None]
        kwds["scalar_code"] = scalar_code
        kwds["vector_code"] = vector_code
        module = self.template.make_module(kwds)
        if len([k for k in module.__dict__ if not k.startswith("_")]):
            return MultiTemplate(module)
        else:
            return autoindent_postfilter(str(module))


class MultiTemplate(Mapping):
    """
    Code generated by a `CodeObjectTemplate` with multiple blocks

    Each block is a string stored as an attribute with the block name. The
    object can also be accessed as a dictionary.
    """

    def __init__(self, module):
        self._templates = {}
        for k, f in module.__dict__.items():
            if not k.startswith("_"):
                s = autoindent_postfilter(str(f()))
                setattr(self, k, s)
                self._templates[k] = s

    def __getitem__(self, item):
        return self._templates[item]

    def __iter__(self):
        return iter(self._templates)

    def __len__(self):
        return len(self._templates)

    def __str__(self):
        s = ""
        for k, v in list(self._templates.items()):
            s += f"{k}:\n"
            s += f"{strip_empty_lines(indent(v))}\n"
        return s

    __repr__ = __str__



"""
Module implementing the C++ "standalone" device.
"""

import inspect
import itertools
import numbers
import os
import shutil
import subprocess
import sys
import tempfile
import time
import zlib
from collections import Counter, defaultdict
from collections.abc import Mapping
from distutils import ccompiler
from hashlib import md5

import numpy as np

from brian2.codegen.codeobject import check_compiler_kwds
from brian2.codegen.cpp_prefs import get_compiler_and_args, get_msvc_env
from brian2.codegen.generators.cpp_generator import c_data_type
from brian2.core.functions import Function
from brian2.core.namespace import get_local_namespace
from brian2.core.network import Network
from brian2.core.preferences import BrianPreference, prefs
from brian2.core.variables import (
    ArrayVariable,
    Constant,
    DynamicArrayVariable,
    Variable,
    VariableView,
)
from brian2.devices.device import Device, all_devices, reset_device, set_device
from brian2.groups.group import Group
from brian2.input import TimedArray
from brian2.parsing.rendering import CPPNodeRenderer
from brian2.synapses.synapses import Synapses
from brian2.units import second
from brian2.units.fundamentalunits import Quantity, fail_for_dimension_mismatch
from brian2.utils.filelock import FileLock
from brian2.utils.filetools import copy_directory, ensure_directory, in_directory
from brian2.utils.logger import get_logger, std_silent
from brian2.utils.stringtools import word_substitute

from .codeobject import CPPStandaloneCodeObject, openmp_pragma

__all__ = []

logger = get_logger(__name__)


# Preferences
prefs.register_preferences(
    "devices.cpp_standalone",
    "C++ standalone preferences ",
    openmp_threads=BrianPreference(
        default=0,
        docs="""
        The number of threads to use if OpenMP is turned on. By default, this value is set to 0 and the C++ code
        is generated without any reference to OpenMP. If greater than 0, then the corresponding number of threads
        are used to launch the simulation.
        """,
    ),
    openmp_spatialneuron_strategy=BrianPreference(
        default=None,
        validator=lambda val: val in [None, "branches", "systems"],
        docs="""
        DEPRECATED. Previously used to chose the strategy to parallelize the
        solution of the three tridiagonal systems for multicompartmental
        neurons. Now, its value is ignored.
        """,
    ),
    make_cmd_unix=BrianPreference(
        default="make",
        docs="""
        The make command used to compile the standalone project. Defaults to the
        standard GNU make commane "make".""",
    ),
    run_cmd_unix=BrianPreference(
        default="./main",
        validator=lambda val: isinstance(val, str) or isinstance(val, list),
        docs="""
        The command used to run the compiled standalone project. Defaults to executing
        the compiled binary with "./main". Must be a single binary as string or a list
        of command arguments (e.g. ["./binary", "--key", "value"]).
        """,
    ),
    extra_make_args_unix=BrianPreference(
        default=["-j"],
        docs="""
        Additional flags to pass to the GNU make command on Linux/OS-X.
        Defaults to "-j" for parallel compilation.""",
    ),
    extra_make_args_windows=BrianPreference(
        default=[],
        docs="""
        Additional flags to pass to the nmake command on Windows. By default, no
        additional flags are passed.
        """,
    ),
    run_environment_variables=BrianPreference(
        default={"LD_BIND_NOW": "1"},
        docs="""
        Dictionary of environment variables and their values that will be set
        during the execution of the standalone code.
        """,
    ),
)


class CPPWriter:
    def __init__(self, project_dir):
        self.project_dir = project_dir
        self.source_files = set()
        self.header_files = set()

    def write(self, filename, contents):
        logger.diagnostic(f"Writing file {filename}:\n{contents}")
        if filename.lower().endswith(".cpp") or filename.lower().endswith(".c"):
            self.source_files.add(filename)
        elif filename.lower().endswith(".h"):
            self.header_files.add(filename)
        elif filename.endswith(".*"):
            self.write(f"{filename[:-1]}cpp", contents.cpp_file)
            self.write(f"{filename[:-1]}h", contents.h_file)
            return
        fullfilename = os.path.join(self.project_dir, filename)
        if os.path.exists(fullfilename):
            with open(fullfilename) as f:
                if f.read() == contents:
                    return
        with open(fullfilename, "w") as f:
            f.write(contents)


def invert_dict(x):
    return {v: k for k, v in x.items()}


class CPPStandaloneDevice(Device):
    """
    The `Device` used for C++ standalone simulations.
    """

    def __init__(self):
        super().__init__()
        #: Dictionary mapping `ArrayVariable` objects to their globally
        #: unique name
        self.arrays = {}
        #: Dictionary mapping `ArrayVariable` objects to their value or to
        #: ``None`` if the value (potentially) depends on executed code. This
        #: mechanism allows to access state variables in standalone mode if
        #: their value is known at run time
        self.array_cache = {}
        #: List of all dynamic arrays
        #: Dictionary mapping `DynamicArrayVariable` objects with 1 dimension to
        #: their globally unique name
        self.dynamic_arrays = {}
        #: Dictionary mapping `DynamicArrayVariable` objects with 2 dimensions
        #: to their globally unique name
        self.dynamic_arrays_2d = {}
        #: List of all arrays to be filled with zeros (list of (var, varname) )
        self.zero_arrays = []
        #: List of all arrays to be filled with numbers (list of
        #: (var, varname, start) tuples
        self.arange_arrays = []
        #: Set of all existing synapses
        self.synapses = set()
        #: Whether the simulation has been run
        self.has_been_run = False

        #: Whether apply_run_args has been called
        self.run_args_applied = False

        #: Whether a run should trigger a build
        self.build_on_run = False

        #: build options
        self.build_options = None

        #: The directory which contains the generated code and results
        self.project_dir = None

        #: The directory which contains the results (relative to `project_dir``)
        self.results_dir = None

        #: Whether to generate profiling information (stored in an instance
        #: variable to be accessible during CodeObject generation)
        self.enable_profiling = False

        #: CodeObjects that use profiling (users can potentially enable
        #: profiling only for a subset of runs)
        self.profiled_codeobjects = []

        #: Dict of all static saved arrays
        self.static_arrays = {}

        #: Names of static arrays used for run_args given as lists of values
        self.run_args_arrays = []

        #: Dict of all TimedArray objects
        self.timed_arrays = {}

        self.code_objects = {}
        self.main_queue = []
        self.runfuncs = {}
        self.networks = set()
        self.static_array_specs = []
        self.report_func = ""

        #: Code lines that have been manually added with `device.insert_code`
        #: Dictionary mapping slot names to lists of lines.
        #: Note that the main slot is handled separately as part of `main_queue`
        self.code_lines = {
            "before_start": [],
            "after_start": [],
            "before_network_run": [],
            "after_network_run": [],
            "before_end": [],
            "after_end": [],
        }

        #: Dictionary storing compile and binary execution times
        self.timers = {"run_binary": None, "compile": {"clean": None, "make": None}}

        self.clocks = set()

        self.extra_compile_args = []
        self.define_macros = []
        self.headers = []
        self.include_dirs = []
        self.library_dirs = []
        self.runtime_library_dirs = []
        self.run_environment_variables = {}
        if sys.platform.startswith("darwin"):
            if "DYLD_LIBRARY_PATH" in os.environ:
                dyld_library_path = f"{os.environ['DYLD_LIBRARY_PATH']}:{os.path.join(sys.prefix, 'lib')}"
            else:
                dyld_library_path = os.path.join(sys.prefix, "lib")
            self.run_environment_variables["DYLD_LIBRARY_PATH"] = dyld_library_path
        self.libraries = []
        if sys.platform == "win32":
            self.libraries += ["advapi32"]
        self.extra_link_args = []
        self.writer = None

    def reinit(self):
        # Remember the build_on_run setting and its options -- important during
        # testing
        build_on_run = self.build_on_run
        build_options = self.build_options
        self.__init__()
        super().reinit()
        self.build_on_run = build_on_run
        self.build_options = build_options

    def spike_queue(self, source_start, source_end):
        return None  # handled differently

    def freeze(self, code, ns):
        # TODO: Remove this function at some point
        logger.warn(
            "The CPPStandaloneDevice.freeze function should no longer "
            "be used, add constant definitions directly to the "
            'code in the "CONSTANTS" section instead.',
            name_suffix="deprecated_freeze_use",
            once=True,
        )
        # this is a bit of a hack, it should be passed to the template somehow
        for k, v in ns.items():
            if isinstance(v, Variable) and v.scalar and v.constant and v.read_only:
                try:
                    v = v.get_value()
                except NotImplementedError:
                    continue
            if isinstance(v, str):
                code = word_substitute(code, {k: v})
            elif isinstance(v, numbers.Number):
                # Use a renderer to correctly transform constants such as True or inf
                renderer = CPPNodeRenderer()
                string_value = renderer.render_expr(repr(v))
                if prefs.core.default_float_dtype == np.float32 and isinstance(
                    v, (float, np.float32, np.float64)
                ):
                    string_value += "f"
                if v < 0:
                    string_value = f"({string_value})"
                code = word_substitute(code, {k: string_value})
            else:
                pass  # don't deal with this object
        return code

    def insert_code(self, slot, code):
        """
        Insert code directly into main.cpp
        """
        if slot == "main":
            self.main_queue.append(("insert_code", code))
        elif slot in self.code_lines:
            self.code_lines[slot].append(code)
        else:
            logger.warn(f"Ignoring device code, unknown slot: {slot}, code: {code}")

    def apply_run_args(self):
        if self.run_args_applied:
            raise RuntimeError(
                "The 'apply_run_args()' function can only be called once."
            )
        self.insert_code("main", "set_from_command_line(args);")
        self.run_args_applied = True

    def static_array(self, name, arr):
        arr = np.atleast_1d(arr)
        assert len(arr), f"length for {name}: {len(arr)}"
        name = f"_static_array_{name}"
        basename = name
        i = 0
        while name in self.static_arrays:
            i += 1
            name = f"{basename}_{str(i)}"
        self.static_arrays[name] = arr.copy()
        return name

    def get_array_name(self, var, access_data=True):
        """
        Return a globally unique name for `var`.

        Parameters
        ----------
        access_data : bool, optional
            For `DynamicArrayVariable` objects, specifying `True` here means the
            name for the underlying data is returned. If specifying `False`,
            the name of object itself is returned (e.g. to allow resizing).
        """
        if isinstance(var, DynamicArrayVariable):
            if access_data:
                return self.arrays[var]
            elif var.ndim == 1:
                return self.dynamic_arrays[var]
            else:
                return self.dynamic_arrays_2d[var]
        elif isinstance(var, ArrayVariable):
            return self.arrays[var]
        else:
            raise TypeError(f"Do not have a name for variable of type {type(var)}.")

    def get_array_filename(self, var, basedir=None):
        """
        Return a file name for a variable.

        Parameters
        ----------
        var : `ArrayVariable`
            The variable to get a filename for.
        basedir : str
            The base directory for the filename, defaults to ``'results'``.
            DEPRECATED: Will raise an error if specified.
        Returns
        -------
        filename : str
            A filename of the form
            ``varname+'_'+str(zlib.crc32(varname))``, where varname
            is the name returned by `get_array_name`.

        Notes
        -----
        The reason that the filename is not simply ``varname`` is
        that this could lead to file names that are not unique in file systems
        that are not case sensitive (e.g. on Windows).
        """
        if basedir is not None:
            raise ValueError("Specifying 'basedir' is no longer supported.")
        varname = self.get_array_name(var, access_data=False)
        return f"{varname}_{str(zlib.crc32(varname.encode('utf-8')))}"

    def add_array(self, var):
        # Note that a dynamic array variable is added to both the arrays and
        # the _dynamic_array dictionary
        if isinstance(var, DynamicArrayVariable):
            # The code below is slightly more complicated than just looking
            # for a unique name as above for static_array, the name has
            # potentially to be unique for more than one dictionary, with
            # different prefixes. This is because dynamic arrays are added to
            # a ``dynamic_arrays`` dictionary (with a `_dynamic` prefix) and to
            # the general ``arrays`` dictionary. We want to make sure that we
            # use the same name in the two dictionaries, not for example
            # ``_dynamic_array_source_name_2`` and ``_array_source_name_1``
            # (this would work fine, but it would make the code harder to read).
            orig_dynamic_name = dynamic_name = (
                f"_dynamic_array_{var.owner.name}_{var.name}"
            )
            orig_array_name = array_name = f"_array_{var.owner.name}_{var.name}"
            suffix = 0

            if var.ndim == 1:
                dynamic_dict = self.dynamic_arrays
            elif var.ndim == 2:
                dynamic_dict = self.dynamic_arrays_2d
            else:
                raise AssertionError(
                    "Did not expect a dynamic array with {var.ndim} dimensions."
                )
            while (
                dynamic_name in dynamic_dict.values()
                or array_name in self.arrays.values()
            ):
                suffix += 1
                dynamic_name = f"{orig_dynamic_name}_{int(suffix)}"
                array_name = f"{orig_array_name}_{int(suffix)}"
            dynamic_dict[var] = dynamic_name
            self.arrays[var] = array_name
        else:
            orig_array_name = array_name = f"_array_{var.owner.name}_{var.name}"
            suffix = 0
            while array_name in self.arrays.values():
                suffix += 1
                array_name = f"{orig_array_name}_{int(suffix)}"
            self.arrays[var] = array_name

    def init_with_zeros(self, var, dtype):
        if isinstance(var, DynamicArrayVariable):
            varname = f"_dynamic{self.arrays[var]}"
        else:
            varname = self.arrays[var]
        self.zero_arrays.append((var, varname))
        self.array_cache[var] = np.zeros(var.size, dtype=dtype)

    def init_with_arange(self, var, start, dtype):
        if isinstance(var, DynamicArrayVariable):
            varname = f"_dynamic{self.arrays[var]}"
        else:
            varname = self.arrays[var]
        self.arange_arrays.append((var, varname, start))
        self.array_cache[var] = np.arange(0, var.size, dtype=dtype) + start

    def fill_with_array(self, var, arr):
        arr = np.asarray(arr)
        if arr.size == 0:
            return  # nothing to do
        array_name = self.get_array_name(var, access_data=False)
        if isinstance(var, DynamicArrayVariable):
            # We can never be sure about the size of a dynamic array, so
            # we can't do correct broadcasting. Therefore, we do not cache
            # them at all for now.
            self.array_cache[var] = None
        else:
            new_arr = np.empty(var.size, dtype=var.dtype)
            new_arr[:] = arr
            self.array_cache[var] = new_arr

        if arr.size == 1:
            if var.size == 1:
                value = CPPNodeRenderer().render_expr(repr(arr.item(0)))
                # For a single assignment, generate a code line instead of storing the array
                self.main_queue.append(("set_by_single_value", (array_name, 0, value)))
            else:
                self.main_queue.append(
                    (
                        "set_by_constant",
                        (array_name, arr.item(), isinstance(var, DynamicArrayVariable)),
                    )
                )
        else:
            # Using the std::vector instead of a pointer to the underlying
            # data for dynamic arrays is fast enough here and it saves us some
            # additional work to set up the pointer
            static_array_name = self.static_array(array_name, arr)
            self.main_queue.append(
                (
                    "set_by_array",
                    (
                        array_name,
                        static_array_name,
                        isinstance(var, DynamicArrayVariable),
                    ),
                )
            )

    def resize(self, var, new_size):
        array_name = self.get_array_name(var, access_data=False)
        self.main_queue.append(("resize_array", (array_name, new_size)))

    def variableview_set_with_index_array(self, variableview, item, value, check_units):
        if isinstance(item, slice) and item == slice(None):
            item = "True"
        value = Quantity(value)

        if (
            isinstance(item, int) or (isinstance(item, np.ndarray) and item.shape == ())
        ) and value.size == 1:
            array_name = self.get_array_name(variableview.variable, access_data=False)
            value_str = CPPNodeRenderer().render_expr(repr(np.asarray(value).item(0)))
            if self.array_cache.get(variableview.variable, None) is not None:
                self.array_cache[variableview.variable][item] = value
            # For a single assignment, generate a code line instead of storing the array
            self.main_queue.append(
                ("set_by_single_value", (array_name, item, value_str))
            )
        # Simple case where we don't have to do any indexing
        elif item == "True" and variableview.index_var in ("_idx", "0"):
            self.fill_with_array(variableview.variable, value)
        else:
            # We have to calculate indices. This will not work for synaptic
            # variables
            try:
                indices = np.asarray(
                    variableview.indexing(item, index_var=variableview.index_var)
                )
            except NotImplementedError:
                raise NotImplementedError(
                    f"Cannot set variable '{variableview.name}' "
                    "this way in standalone, try using "
                    "string expressions."
                )
            # Using the std::vector instead of a pointer to the underlying
            # data for dynamic arrays is fast enough here and it saves us some
            # additional work to set up the pointer
            arrayname = self.get_array_name(variableview.variable, access_data=False)
            if indices.shape != () and (
                value.shape == () or (value.size == 1 and indices.size > 1)
            ):
                value = np.repeat(value, indices.size)
            elif value.shape != indices.shape and len(value) != len(indices):
                raise ValueError(
                    "Provided values do not match the size "
                    "of the indices, "
                    f"{len(value)} != len(indices)."
                )

            staticarrayname_index = self.static_array(f"_index_{arrayname}", indices)
            staticarrayname_value = self.static_array(f"_value_{arrayname}", value)
            self.array_cache[variableview.variable] = None
            self.main_queue.append(
                (
                    "set_array_by_array",
                    (arrayname, staticarrayname_index, staticarrayname_value),
                )
            )

    def get_value(self, var, access_data=True):
        # Usually, we cannot retrieve the values of state variables in
        # standalone scripts since their values might depend on the evaluation
        # of expressions at runtime. For some variables we do know the value
        # however (values that have been set with explicit values and not
        # changed in code objects)
        if self.array_cache.get(var, None) is not None:
            return self.array_cache[var]
        else:
            # After the network has been run, we can retrieve the values from
            # disk
            if self.has_been_run:
                dtype = var.dtype
                fname = os.path.join(self.results_dir, self.get_array_filename(var))
                with open(fname, "rb") as f:
                    data = np.fromfile(f, dtype=dtype)
                # This is a bit of an heuristic, but our 2d dynamic arrays are
                # only expanding in one dimension, we assume here that the
                # other dimension has size 0 at the beginning
                if isinstance(var.size, tuple) and len(var.size) == 2:
                    if var.size[0] * var.size[1] == len(data):
                        size = var.size
                    elif var.size[0] == 0:
                        size = (len(data) // var.size[1], var.size[1])
                    elif var.size[1] == 0:
                        size = (var.size[0], len(data) // var.size[0])
                    else:
                        raise IndexError(
                            "Do not now how to deal with 2d "
                            f"array of size {var.size!s}, the array on "
                            f"disk has length {len(data)}."
                        )

                    var.size = size
                    return data.reshape(var.size)
                var.size = len(data)
                return data
            raise NotImplementedError(
                "Cannot retrieve the values of state "
                "variables in standalone code before the "
                "simulation has been run."
            )

    def variableview_get_subexpression_with_index_array(
        self, variableview, item, run_namespace=None
    ):
        if not self.has_been_run:
            raise NotImplementedError(
                "Cannot retrieve the values of state "
                "variables in standalone code before the "
                "simulation has been run."
            )
        # Temporarily switch to the runtime device to evaluate the subexpression
        # (based on the values stored on disk)
        set_device("runtime")
        result = VariableView.get_subexpression_with_index_array(
            variableview, item, run_namespace=run_namespace
        )
        reset_device()
        return result

    def variableview_get_with_expression(self, variableview, code, run_namespace=None):
        raise NotImplementedError(
            "Cannot retrieve the values of state "
            "variables with string expressions in "
            "standalone scripts."
        )

    def code_object_class(self, codeobj_class=None, fallback_pref=None):
        """
        Return `CodeObject` class (either `CPPStandaloneCodeObject` class or input)

        Parameters
        ----------
        codeobj_class : a `CodeObject` class, optional
            If this is keyword is set to None or no arguments are given, this method will return
            the default (`CPPStandaloneCodeObject` class).
        fallback_pref : str, optional
            For the cpp_standalone device this option is ignored.

        Returns
        -------
        codeobj_class : class
            The `CodeObject` class that should be used
        """
        # Ignore the requested pref (used for optimization in runtime)
        if codeobj_class is None:
            return CPPStandaloneCodeObject
        else:
            return codeobj_class

    def code_object(
        self,
        owner,
        name,
        abstract_code,
        variables,
        template_name,
        variable_indices,
        codeobj_class=None,
        template_kwds=None,
        override_conditional_write=None,
        compiler_kwds=None,
    ):
        if compiler_kwds is None:
            compiler_kwds = {}
        check_compiler_kwds(
            compiler_kwds,
            [
                "headers",
                "sources",
                "define_macros",
                "libraries",
                "include_dirs",
                "library_dirs",
                "runtime_library_dirs",
            ],
            "C++ standalone",
        )
        if template_kwds is None:
            template_kwds = dict()
        else:
            template_kwds = dict(template_kwds)
        # In standalone mode, the only place where we use additional header
        # files is by inserting them into the template
        codeobj_headers = compiler_kwds.get("headers", [])
        template_kwds["user_headers"] = (
            self.headers + prefs["codegen.cpp.headers"] + codeobj_headers
        )
        template_kwds["profiled"] = self.enable_profiling

        do_not_invalidate = set()
        if template_name == "synapses_create_array":
            cache = self.array_cache
            if (
                cache[variables["N"]] is None
            ):  # synapses have been previously created with code
                # Nothing we can do
                logger.debug(
                    f"Synapses for '{owner.name}' have previously been created with "
                    "code, we therefore cannot cache the synapses created with arrays "
                    f"via '{name}'",
                    name_suffix="code_created_synapses_exist",
                )
            else:  # first time we create synapses, or all previous connect calls were with arrays
                cache[variables["N"]][0] += variables["sources"].size
                do_not_invalidate.add(variables["N"])
                for var, value in [
                    (
                        variables["_synaptic_pre"],
                        variables["sources"].get_value()
                        + variables["_source_offset"].get_value(),
                    ),
                    (
                        variables["_synaptic_post"],
                        variables["targets"].get_value()
                        + variables["_target_offset"].get_value(),
                    ),
                ]:
                    cache[var] = np.append(
                        cache.get(var, np.empty(0, dtype=int)), value
                    )
                    do_not_invalidate.add(var)

        codeobj = super().code_object(
            owner,
            name,
            abstract_code,
            variables,
            template_name,
            variable_indices,
            codeobj_class=codeobj_class,
            template_kwds=template_kwds,
            override_conditional_write=override_conditional_write,
            compiler_kwds=compiler_kwds,
        )
        self.code_objects[codeobj.name] = codeobj
        if self.enable_profiling:
            self.profiled_codeobjects.append(codeobj.name)

        for var in codeobj.variables.values():
            if isinstance(var, TimedArray):
                self.timed_arrays[var] = var.name
        # We mark all writeable (i.e. not read-only) variables used by the code
        # as "dirty" to avoid that the cache contains incorrect values. This
        # might remove a number of variables from the cache unnecessarily,
        # since many variables are only read in the code.
        # On the other hand, there are also *read-only* variables that can be
        # changed by code (the "read-only" attribute only refers to the user
        # being able to change values directly). For example, synapse creation
        # write source and target indices, and monitors write the shared values.
        # To correctly mark these values as changed, templates can include a
        # "WRITES_TO_READ_ONLY_VARIABLES" comment, stating the name of the
        # changed variables. For a monitor, this would for example state that
        # the number of recorded values "N" changes. For the recorded variables,
        # however, this information cannot be included in the template because
        # it is up to the user to define which variables are recorded. For such
        # cases, the "owner" object (e.g. a SpikeMonitor) can define a
        # "written_readonly_vars" attribute, storing a set of `Variable` objects
        # that will be changed by the owner's code objects.
        template = getattr(codeobj.templater, template_name)
        written_readonly_vars = {
            codeobj.variables[varname] for varname in template.writes_read_only
        } | getattr(owner, "written_readonly_vars", set())
        for var in codeobj.variables.values():
            if (
                isinstance(var, ArrayVariable)
                and var not in do_not_invalidate
                and (not var.read_only or var in written_readonly_vars)
            ):
                self.array_cache[var] = None

        return codeobj

    def check_openmp_compatible(self, nb_threads):
        if nb_threads > 0:
            logger.warn(
                "OpenMP code is not yet well tested, and may be inaccurate.",
                "openmp",
                once=True,
            )
            logger.diagnostic(f"Using OpenMP with {int(nb_threads)} threads ")
            if prefs.devices.cpp_standalone.openmp_spatialneuron_strategy is not None:
                logger.warn(
                    "The devices.cpp_standalone.openmp_spatialneuron_strategy "
                    "preference is no longer used and will be removed in "
                    "future versions of Brian.",
                    "openmp_spatialneuron_strategy",
                    once=True,
                )

    def generate_objects_source(
        self,
        writer,
        arange_arrays,
        synapses,
        static_array_specs,
        networks,
        timed_arrays,
    ):
        arr_tmp = self.code_object_class().templater.objects(
            None,
            None,
            array_specs=self.arrays,
            dynamic_array_specs=self.dynamic_arrays,
            dynamic_array_2d_specs=self.dynamic_arrays_2d,
            zero_arrays=self.zero_arrays,
            arange_arrays=arange_arrays,
            synapses=synapses,
            clocks=self.clocks,
            static_array_specs=static_array_specs,
            networks=networks,
            get_array_filename=self.get_array_filename,
            get_array_name=self.get_array_name,
            profiled_codeobjects=self.profiled_codeobjects,
            code_objects=list(self.code_objects.values()),
            timed_arrays=timed_arrays,
        )
        writer.write("objects.*", arr_tmp)

    def generate_main_source(self, writer):
        main_lines = []
        procedures = [("", main_lines)]
        runfuncs = {}
        for func, args in self.main_queue:
            if func == "before_run_code_object":
                (codeobj,) = args
                main_lines.append(f"_before_run_{codeobj.name}();")
            elif func == "run_code_object":
                (codeobj,) = args
                main_lines.append(f"_run_{codeobj.name}();")
            elif func == "after_run_code_object":
                (codeobj,) = args
                main_lines.append(f"_after_run_{codeobj.name}();")
            elif func == "run_network":
                net, netcode = args
                main_lines.extend(netcode)
            elif func == "set_by_constant":
                arrayname, value, is_dynamic = args
                size_str = f"{arrayname}.size()" if is_dynamic else f"_num_{arrayname}"
                code = f"""
                {openmp_pragma('static')}
                for(int i=0; i<{size_str}; i++)
                {{
                    {arrayname}[i] = {CPPNodeRenderer().render_expr(repr(value))};
                }}
                """
                main_lines.extend(code.split("\n"))
            elif func == "set_by_array":
                arrayname, staticarrayname, is_dynamic = args
                size_str = f"{arrayname}.size()" if is_dynamic else f"_num_{arrayname}"
                code = f"""
                {openmp_pragma('static')}
                for(int i=0; i<{size_str}; i++)
                {{
                    {arrayname}[i] = {staticarrayname}[i];
                }}
                """
                main_lines.extend(code.split("\n"))
            elif func == "set_by_single_value":
                arrayname, item, value = args
                code = f"{arrayname}[{item}] = {value};"
                main_lines.extend([code])
            elif func == "set_array_by_array":
                arrayname, staticarrayname_index, staticarrayname_value = args
                code = f"""
                {openmp_pragma('static')}
                for(int i=0; i<_num_{staticarrayname_index}; i++)
                {{
                    {arrayname}[{staticarrayname_index}[i]] = {staticarrayname_value}[i];
                }}
                """
                main_lines.extend(code.split("\n"))
            elif func == "resize_array":
                array_name, new_size = args
                main_lines.append(f"{array_name}.resize({new_size});")
            elif func == "insert_code":
                main_lines.append(args)
            elif func == "start_run_func":
                name, include_in_parent = args
                if include_in_parent:
                    main_lines.append(f"{name}();")
                main_lines = []
                procedures.append((name, main_lines))
            elif func == "end_run_func":
                name, include_in_parent = args
                name, main_lines = procedures.pop(-1)
                runfuncs[name] = main_lines
                name, main_lines = procedures[-1]
            elif func == "seed":
                seed = args
                nb_threads = prefs.devices.cpp_standalone.openmp_threads
                if nb_threads == 0:  # no OpenMP
                    nb_threads = 1
                main_lines.append(f"for (int _i=0; _i<{nb_threads}; _i++)")
                if seed is None:  # random
                    main_lines.append("    brian::_random_generators[_i].seed();")
                else:
                    main_lines.append(
                        f"    brian::_random_generators[_i].seed({seed!r}L + _i);"
                    )
            else:
                raise NotImplementedError(f"Unknown main queue function type {func}")

        self.runfuncs = runfuncs

        # generate the finalisations
        for codeobj in self.code_objects.values():
            if hasattr(codeobj.code, "main_finalise"):
                main_lines.append(codeobj.code.main_finalise)

        user_headers = self.headers + prefs["codegen.cpp.headers"]
        main_tmp = self.code_object_class().templater.main(
            None,
            None,
            main_lines=main_lines,
            code_lines=self.code_lines,
            code_objects=list(self.code_objects.values()),
            report_func=self.report_func,
            dt=float(self.defaultclock.dt),
            user_headers=user_headers,
        )
        writer.write("main.cpp", main_tmp)

    def generate_codeobj_source(self, writer):
        # Generate data for non-constant values
        renderer = CPPNodeRenderer()
        code_object_defs = defaultdict(list)
        for codeobj in self.code_objects.values():
            lines = []
            for k, v in codeobj.variables.items():
                if isinstance(v, ArrayVariable):
                    try:
                        if isinstance(v, DynamicArrayVariable):
                            if v.ndim == 1:
                                dyn_array_name = self.dynamic_arrays[v]
                                array_name = self.arrays[v]
                                c_type = c_data_type(v.dtype)
                                line = (
                                    f"{c_type}* const {array_name} ="
                                    f" {dyn_array_name}.empty()? 0 :"
                                    f" &{dyn_array_name}[0];"
                                )
                                lines.append(line)
                                line = (
                                    f"const size_t _num{k} = {dyn_array_name}.size();"
                                )
                                lines.append(line)
                        else:
                            lines.append(f"const size_t _num{k} = {v.size};")
                    except TypeError:
                        pass
                elif isinstance(v, Constant):
                    value = renderer.render_expr(repr(v.value))
                    c_type = c_data_type(v.dtype)
                    line = f"const {c_type} {k} = {value};"
                    lines.append(line)
            for line in lines:
                # Sometimes an array is referred to by to different keys in our
                # dictionary -- make sure to never add a line twice
                if line not in code_object_defs[codeobj.name]:
                    code_object_defs[codeobj.name].append(line)

        # Generate the code objects
        for codeobj in self.code_objects.values():
            # Before/after run code
            for block in codeobj.before_after_blocks:
                cpp_code = getattr(codeobj.code, f"{block}_cpp_file")
                cpp_code = cpp_code.replace(
                    "%CONSTANTS%", "\n".join(code_object_defs[codeobj.name])
                )
                h_code = getattr(codeobj.code, f"{block}_h_file")
                writer.write(f"code_objects/{block}_{codeobj.name}.cpp", cpp_code)
                writer.write(f"code_objects/{block}_{codeobj.name}.h", h_code)

            # Main code
            code = codeobj.code.cpp_file
            code = code.replace(
                "%CONSTANTS%", "\n".join(code_object_defs[codeobj.name])
            )

            writer.write(f"code_objects/{codeobj.name}.cpp", code)
            writer.write(f"code_objects/{codeobj.name}.h", codeobj.code.h_file)

    def generate_network_source(self, writer, compiler):
        maximum_run_time = self._maximum_run_time
        if maximum_run_time is not None:
            maximum_run_time = float(maximum_run_time)
        network_tmp = self.code_object_class().templater.network(
            None, None, maximum_run_time=maximum_run_time
        )
        writer.write("network.*", network_tmp)

    def generate_synapses_classes_source(self, writer):
        synapses_classes_tmp = self.code_object_class().templater.synapses_classes(
            None, None
        )
        writer.write("synapses_classes.*", synapses_classes_tmp)

    def generate_run_source(self, writer):
        run_tmp = self.code_object_class().templater.run(
            None,
            None,
            run_funcs=self.runfuncs,
            code_objects=list(self.code_objects.values()),
            user_headers=self.headers,
            array_specs=self.arrays,
            clocks=self.clocks,
        )
        writer.write("run.*", run_tmp)

    def generate_makefile(
        self, writer, compiler, compiler_flags, linker_flags, nb_threads, debug
    ):
        if compiler == "msvc":
            if nb_threads > 1:
                openmp_flag = "/openmp"
            else:
                openmp_flag = ""
            if debug:
                compiler_debug_flags = "/DEBUG /DDEBUG"
                linker_debug_flags = "/DEBUG"
            else:
                compiler_debug_flags = ""
                linker_debug_flags = ""
            # Generate the visual studio makefile
            source_bases = [
                fname.replace(".cpp", "").replace(".c", "").replace("/", "\\")
                for fname in sorted(writer.source_files)
            ]
            win_makefile_tmp = self.code_object_class().templater.win_makefile(
                None,
                None,
                source_files=sorted(writer.source_files),
                source_bases=source_bases,
                compiler_flags=compiler_flags,
                compiler_debug_flags=compiler_debug_flags,
                linker_flags=linker_flags,
                linker_debug_flags=linker_debug_flags,
                openmp_flag=openmp_flag,
            )
            writer.write("win_makefile", win_makefile_tmp)
            # write the list of sources
            source_list = " ".join(source_bases)
            source_list_fname = os.path.join(self.project_dir, "sourcefiles.txt")
            if os.path.exists(source_list_fname):
                with open(source_list_fname) as f:
                    if f.read() == source_list:
                        return
            with open(source_list_fname, "w") as f:
                f.write(source_list)
        else:
            # Generate the makefile
            if os.name == "nt":
                rm_cmd = "del *.o /s\n\tdel main.exe $(DEPS)"
            else:
                rm_cmd = "rm $(OBJS) $(PROGRAM) $(DEPS)"
            if debug:
                compiler_debug_flags = "-g -DDEBUG"
                linker_debug_flags = "-g"
            else:
                compiler_debug_flags = ""
                linker_debug_flags = ""
            makefile_tmp = self.code_object_class().templater.makefile(
                None,
                None,
                source_files=" ".join(sorted(writer.source_files)),
                header_files=" ".join(sorted(writer.header_files)),
                compiler_flags=compiler_flags,
                compiler_debug_flags=compiler_debug_flags,
                linker_debug_flags=linker_debug_flags,
                linker_flags=linker_flags,
                rm_cmd=rm_cmd,
            )
            writer.write("makefile", makefile_tmp)

    def copy_source_files(self, writer, directory):
        # Copy the brianlibdirectory
        brianlib_dir = os.path.join(
            os.path.split(inspect.getsourcefile(CPPStandaloneCodeObject))[0], "brianlib"
        )
        brianlib_files = copy_directory(
            brianlib_dir, os.path.join(directory, "brianlib")
        )
        for file in brianlib_files:
            if file.lower().endswith(".cpp"):
                writer.source_files.add(f"brianlib/{file}")
            elif file.lower().endswith(".h"):
                writer.header_files.add(f"brianlib/{file}")

        # Copy the CSpikeQueue implementation
        shutil.copy2(
            os.path.join(
                os.path.split(inspect.getsourcefile(Synapses))[0], "cspikequeue.cpp"
            ),
            os.path.join(directory, "brianlib", "spikequeue.h"),
        )
        shutil.copy2(
            os.path.join(
                os.path.split(inspect.getsourcefile(Synapses))[0], "stdint_compat.h"
            ),
            os.path.join(directory, "brianlib", "stdint_compat.h"),
        )

    def _insert_func_namespace(self, func, code_object, namespace):
        impl = func.implementations[self.code_object_class()]
        func_namespace = impl.get_namespace(code_object.owner)
        if func_namespace is not None:
            namespace.update(func_namespace)
        if impl.dependencies is not None:
            for dep in impl.dependencies.values():
                self._insert_func_namespace(dep, code_object, namespace)

    def write_static_arrays(self, directory):
        # Write Function namespaces as static arrays
        for code_object in self.code_objects.values():
            for var in code_object.variables.values():
                if isinstance(var, Function):
                    self._insert_func_namespace(var, code_object, self.static_arrays)

        logger.diagnostic(f"static arrays: {str(sorted(self.static_arrays.keys()))}")

        static_array_specs = []
        for name, arr in sorted(self.static_arrays.items()):
            arr.tofile(os.path.join(directory, "static_arrays", name))
            static_array_specs.append((name, c_data_type(arr.dtype), arr.size, name))
        self.static_array_specs = static_array_specs

    def compile_source(self, directory, compiler, debug, clean):
        with in_directory(directory):
            if compiler == "msvc":
                msvc_env, vcvars_cmd = get_msvc_env()
                make_cmd = "nmake /f win_makefile"
                make_args = " ".join(
                    prefs.devices.cpp_standalone.extra_make_args_windows
                )
                if os.path.exists("winmake.log"):
                    os.remove("winmake.log")
                if vcvars_cmd:
                    with open("winmake.log", "w") as f:
                        f.write(f"{vcvars_cmd}\n")
                else:
                    with open("winmake.log", "w") as f:
                        f.write("MSVC environment: \n")
                        for key, value in msvc_env.items():
                            f.write(f"{key}={value}\n")
                with std_silent(debug):
                    if vcvars_cmd:
                        if clean:
                            start_time = time.time()
                            os.system(
                                f"{vcvars_cmd} >>winmake.log 2>&1 && {make_cmd} clean >"
                                " NUL 2>&1"
                            )
                            self.timers["compile"]["clean"] = time.time() - start_time
                        start_time = time.time()
                        x = os.system(
                            f"{vcvars_cmd} >>winmake.log 2>&1 &&"
                            f" {make_cmd} {make_args}>>winmake.log 2>&1"
                        )
                        self.timers["compile"]["make"] = time.time() - start_time
                    else:
                        os.environ.update(msvc_env)
                        if clean:
                            start_time = time.time()
                            os.system(f"{make_cmd} clean > NUL 2>&1")
                            self.timers["compile"]["clean"] = time.time() - start_time
                        start_time = time.time()
                        x = os.system(f"{make_cmd} {make_args}>>winmake.log 2>&1")
                        self.timers["compile"]["make"] = time.time() - start_time

                    if x != 0:
                        error_message = f"Project compilation failed (error code: {x}), consider having a look at 'winmake.log'."
                        if not clean:
                            error_message += (
                                " Consider running with "
                                '"clean=True" to force a complete '
                                "rebuild."
                            )
                        raise RuntimeError(error_message)
            else:
                with std_silent(debug):
                    if clean:
                        start_time = time.time()
                        os.system("make clean >/dev/null 2>&1")
                        self.timers["compile"]["clean"] = time.time() - start_time
                    make_cmd = prefs.devices.cpp_standalone.make_cmd_unix
                    make_args = " ".join(
                        prefs.devices.cpp_standalone.extra_make_args_unix
                    )
                    start_time = time.time()
                    x = os.system(f"{make_cmd} {make_args}")
                    self.timers["compile"]["make"] = time.time() - start_time
                    if x != 0:
                        error_message = (
                            "Project compilation failed (error code: %u)." % x
                        )
                        if not clean:
                            error_message += (
                                " Consider running with "
                                '"clean=True" to force a complete '
                                "rebuild."
                            )
                        raise RuntimeError(error_message)

    def seed(self, seed=None):
        """
        Set the seed for the random number generator.

        Parameters
        ----------
        seed : int, optional
            The seed value for the random number generator, or ``None`` (the
            default) to set a random seed.
        """
        self.main_queue.append(("seed", seed))

    def run(
        self, directory=None, results_directory=None, with_output=True, run_args=None
    ):
        if directory is None:
            directory = self.project_dir
        if results_directory is None:
            results_directory = self.results_dir
        else:
            if os.path.isabs(results_directory):
                raise TypeError(
                    "The 'results_directory' argument needs to be a relative path but"
                    f" was '{results_directory}'."
                )
            # Translate path to absolute path which ends with /
            self.results_dir = os.path.join(
                os.path.abspath(os.path.join(directory, results_directory)), ""
            )
        ensure_directory(self.results_dir)

        self.run_args_arrays.clear()  # forget about arrays from previous runs
        if run_args is None:
            run_args = []
        elif isinstance(run_args, Mapping):
            list_rep = []
            for key, value in run_args.items():
                if isinstance(key, VariableView):
                    (
                        name,
                        string_value,
                        value_name,
                        value_ar,
                    ) = self._prepare_variableview_run_arg(key, value)
                elif isinstance(key, TimedArray):
                    (
                        name,
                        string_value,
                        value_name,
                        value_ar,
                    ) = self._prepare_timed_array_run_arg(key, value)
                else:
                    raise TypeError(
                        "The keys for 'run_args' need to be 'VariableView' objects,"
                        " i.e. attributes of groups such as 'neurongroup.v', or a"
                        f" 'TimedArray'. Key has type '{type(key)}' instead."
                    )
                if value_name:
                    fname = os.path.join(self.project_dir, "static_arrays", value_name)
                    # Make sure processes trying to write the same file don't clash
                    with FileLock(fname + ".lock"):
                        if not os.path.exists(fname):
                            value_ar.tofile(fname)
                    self.run_args_arrays.append(value_name)
                list_rep.append(f"{name}={string_value}")

            run_args = list_rep

        # Invalidate array cache for all variables set on the command line
        for arg in run_args:
            s = arg.split("=")
            if len(s) == 2:
                for var in self.array_cache:
                    if (
                        hasattr(var.owner, "name")
                        and var.owner.name + "." + var.name == s[0]
                    ):
                        self.array_cache[var] = None
        run_args = ["--results_dir", self.results_dir] + run_args
        # Invalidate the cached end time of the clock and network, to deal with stopped simulations
        for clock in self.clocks:
            self.array_cache[clock.variables["t"]] = None

        with in_directory(directory):
            # Set environment variables

            for key, value in itertools.chain(
                prefs["devices.cpp_standalone.run_environment_variables"].items(),
                self.run_environment_variables.items(),
            ):
                if key in os.environ and os.environ[key] != value:
                    logger.info(
                        f'Overwriting environment variable "{key}"',
                        name_suffix="overwritten_env_var",
                        once=True,
                    )
                os.environ[key] = value
            if not with_output:
                stdout = open(os.path.join(self.results_dir, "stdout.txt"), "w")
            else:
                stdout = None
            if os.name == "nt":
                start_time = time.time()
                Network._globally_running = True
                x = subprocess.call(["main"] + run_args, stdout=stdout)
                self.timers["run_binary"] = time.time() - start_time
                Network._globally_running = False
            else:
                run_cmd = prefs.devices.cpp_standalone.run_cmd_unix
                if isinstance(run_cmd, str):
                    run_cmd = [run_cmd]
                start_time = time.time()
                Network._globally_running = True
                x = subprocess.call(run_cmd + run_args, stdout=stdout)
                self.timers["run_binary"] = time.time() - start_time
                Network._globally_running = False
            if stdout is not None:
                stdout.close()
            if x:
                stdout_fname = os.path.join(self.results_dir, "stdout.txt")
                if os.path.exists(stdout_fname):
                    with open(stdout_fname) as f:
                        print(f.read())
                raise RuntimeError(
                    "Project run failed (project directory:"
                    f" {os.path.abspath(directory)})"
                )
            self.has_been_run = True
            run_info_fname = os.path.join(self.results_dir, "last_run_info.txt")
            if os.path.isfile(run_info_fname):
                with open(run_info_fname) as f:
                    last_run_info = f.read()
                run_time, completed_fraction = last_run_info.split()
                self._last_run_time = float(run_time)
                self._last_run_completed_fraction = float(completed_fraction)

        # Make sure that integration did not create NaN or very large values
        owners = [var.owner for var in self.arrays]
        # We don't want to check the same owner twice but var.owner is a
        # weakproxy which we can't put into a set. We therefore store the name
        # of all objects we already checked. Furthermore, under some specific
        # instances a variable might have been created whose owner no longer
        # exists (e.g. a `_sub_idx` variable for a subgroup) -- we ignore the
        # resulting reference error.
        already_checked = set()
        for owner in owners:
            try:
                if not hasattr(owner, "name") or owner.name in already_checked:
                    continue
                if isinstance(owner, Group):
                    owner._check_for_invalid_states()
                    already_checked.add(owner.name)
            except ReferenceError:
                pass

    def _prepare_variableview_run_arg(self, key, value):
        fail_for_dimension_mismatch(key.dim, value)  # TODO: Give name of variable
        value_ar = np.asarray(value, dtype=key.dtype)
        if value_ar.ndim == 0 or value_ar.size == 1:
            # single value, give value directly on command line
            string_value = repr(value_ar.item())
            value_name = None
        else:
            if value_ar.ndim != 1 or (
                not key.variable.dynamic and value_ar.size != key.shape[0]
            ):
                raise TypeError(
                    "Incorrect size for variable"
                    f" '{key.group_name}.{key.name}'. Shape {key.shape} ≠"
                    f" {value_ar.shape}."
                )
            value_name = (
                f"init_{key.group_name}_{key.name}_{md5(value_ar.data).hexdigest()}.dat"
            )
            string_value = os.path.join("static_arrays", value_name)
        name = f"{key.group_name}.{key.name}"
        return name, string_value, value_name, value_ar

    def _prepare_timed_array_run_arg(self, key, value):
        fail_for_dimension_mismatch(key.dim, value)  # TODO: Give name of variable
        value_ar = np.asarray(value, dtype=key.values.dtype)
        if value_ar.ndim == 0 or value_ar.size == 1:
            # single value, give value directly on command line
            string_value = repr(value_ar.item())
            value_name = None
        elif value_ar.shape == key.values.shape:
            value_name = f"init_{key.name}_values_{md5(value_ar.data).hexdigest()}.dat"
            string_value = os.path.join("static_arrays", value_name)
        else:
            raise TypeError(
                "Incorrect size for variable"
                f" '{key.name}.values'. Shape {key.values.shape} ≠"
                f" {value_ar.shape}."
            )
        name = f"{key.name}.values"
        return name, string_value, value_name, value_ar

    def build(
        self,
        directory="output",
        results_directory="results",
        compile=True,
        run=True,
        debug=False,
        clean=False,
        with_output=True,
        additional_source_files=None,
        run_args=None,
        direct_call=True,
        **kwds,
    ):
        """
        Build the project

        TODO: more details

        Parameters
        ----------
        directory : str, optional
            The output directory to write the project to, any existing files
            will be overwritten. If the given directory name is ``None``, then
            a temporary directory will be used (used in the test suite to avoid
            problems when running several tests in parallel). Defaults to
            ``'output'``.
        compile : bool, optional
            Whether or not to attempt to compile the project. Defaults to
            ``True``.
        run : bool, optional
            Whether or not to attempt to run the built project if it
            successfully builds. Defaults to ``True``.
        debug : bool, optional
            Whether to compile in debug mode. Defaults to ``False``.
        with_output : bool, optional
            Whether or not to show the ``stdout`` of the built program when run.
            Output will be shown in case of compilation or runtime error.
            Defaults to ``True``.
        clean : bool, optional
            Whether or not to clean the project before building. Defaults to
            ``False``.
        additional_source_files : list of str, optional
            A list of additional ``.cpp`` files to include in the build.
        direct_call : bool, optional
            Whether this function was called directly. Is used internally to
            distinguish an automatic build due to the ``build_on_run`` option
            from a manual ``device.build`` call.
        """
        if self.build_on_run and direct_call:
            raise RuntimeError(
                "You used set_device with build_on_run=True "
                "(the default option), which will automatically "
                "build the simulation at the first encountered "
                "run call - do not call device.build manually "
                "in this case. If you want to call it manually, "
                "e.g. because you have multiple run calls, use "
                "set_device with build_on_run=False."
            )
        if self.has_been_run:
            raise RuntimeError(
                "The network has already been built and run "
                "before. To build several simulations in "
                'the same script, call "device.reinit()" '
                'and "device.activate()". Note that you '
                "will have to set build options (e.g. the "
                "directory) and defaultclock.dt again."
            )
        renames = {
            "project_dir": "directory",
            "compile_project": "compile",
            "run_project": "run",
        }
        if len(kwds):
            msg = ""
            for kwd in kwds:
                if kwd in renames:
                    msg += (
                        f"Keyword argument '{kwd}' has been renamed to "
                        f"'{renames[kwd]}'. "
                    )
                else:
                    msg += f"Unknown keyword argument '{kwd}'. "
            raise TypeError(msg)

        if additional_source_files is None:
            additional_source_files = []
        if run_args is None:
            run_args = []
        if directory is None:
            directory = tempfile.mkdtemp(prefix="brian_standalone_")
        self.project_dir = directory
        ensure_directory(directory)
        if os.path.isabs(results_directory):
            raise TypeError(
                "The 'results_directory' argument needs to be a relative path but was "
                f"'{results_directory}'."
            )
        # Translate path to absolute path which ends with /
        self.results_dir = os.path.join(
            os.path.abspath(os.path.join(directory, results_directory)), ""
        )

        # Determine compiler flags and directories
        compiler, default_extra_compile_args = get_compiler_and_args()
        extra_compile_args = self.extra_compile_args + default_extra_compile_args
        extra_link_args = self.extra_link_args + prefs["codegen.cpp.extra_link_args"]

        codeobj_define_macros = [
            macro
            for codeobj in self.code_objects.values()
            for macro in codeobj.compiler_kwds.get("define_macros", [])
        ]
        define_macros = (
            self.define_macros
            + prefs["codegen.cpp.define_macros"]
            + codeobj_define_macros
        )

        codeobj_include_dirs = [
            include_dir
            for codeobj in self.code_objects.values()
            for include_dir in codeobj.compiler_kwds.get("include_dirs", [])
        ]
        include_dirs = (
            self.include_dirs + prefs["codegen.cpp.include_dirs"] + codeobj_include_dirs
        )

        codeobj_library_dirs = [
            library_dir
            for codeobj in self.code_objects.values()
            for library_dir in codeobj.compiler_kwds.get("library_dirs", [])
        ]
        library_dirs = (
            self.library_dirs + prefs["codegen.cpp.library_dirs"] + codeobj_library_dirs
        )

        codeobj_runtime_dirs = [
            runtime_dir
            for codeobj in self.code_objects.values()
            for runtime_dir in codeobj.compiler_kwds.get("runtime_library_dirs", [])
        ]
        runtime_library_dirs = (
            self.runtime_library_dirs
            + prefs["codegen.cpp.runtime_library_dirs"]
            + codeobj_runtime_dirs
        )

        codeobj_libraries = [
            library
            for codeobj in self.code_objects.values()
            for library in codeobj.compiler_kwds.get("libraries", [])
        ]
        libraries = self.libraries + prefs["codegen.cpp.libraries"] + codeobj_libraries

        compiler_obj = ccompiler.new_compiler(compiler=compiler)

        # Distutils does not use the shell, so it does not need to quote filenames/paths
        # Since we include the compiler flags in the makefile, we need to quote them
        include_dirs = [f'"{include_dir}"' for include_dir in include_dirs]
        library_dirs = [f'"{library_dir}"' for library_dir in library_dirs]
        runtime_library_dirs = [
            f'"{runtime_dir}"' for runtime_dir in runtime_library_dirs
        ]

        compiler_flags = (
            ccompiler.gen_preprocess_options(define_macros, include_dirs)
            + extra_compile_args
        )

        linker_flags = (
            ccompiler.gen_lib_options(
                compiler_obj,
                library_dirs=library_dirs,
                runtime_library_dirs=runtime_library_dirs,
                libraries=libraries,
            )
            + extra_link_args
        )

        codeobj_source_files = [
            source_file
            for codeobj in self.code_objects.values()
            for source_file in codeobj.compiler_kwds.get("sources", [])
        ]
        additional_source_files += codeobj_source_files

        for d in ["code_objects", "results", "static_arrays"]:
            ensure_directory(os.path.join(directory, d))

        self.writer = CPPWriter(directory)

        # Get the number of threads if specified in an openmp context
        nb_threads = prefs.devices.cpp_standalone.openmp_threads
        # If the number is negative, we need to throw an error
        if nb_threads < 0:
            raise ValueError("The number of OpenMP threads can not be negative !")

        logger.diagnostic(
            "Writing C++ standalone project to directory "
            f"'{os.path.normpath(directory)}'."
        )

        self.check_openmp_compatible(nb_threads)

        self.write_static_arrays(directory)

        # Check that all names are globally unique
        names = [obj.name for net in self.networks for obj in net.sorted_objects]
        non_unique_names = [name for name, count in Counter(names).items() if count > 1]
        if len(non_unique_names):
            formatted_names = ", ".join(f"'{name}'" for name in non_unique_names)
            raise ValueError(
                "All objects need to have unique names in "
                "standalone mode, the following name(s) were used "
                f"more than once: {formatted_names}"
            )

        self.generate_objects_source(
            self.writer,
            self.arange_arrays,
            self.synapses,
            self.static_array_specs,
            self.networks,
            self.timed_arrays,
        )
        self.generate_main_source(self.writer)
        self.generate_codeobj_source(self.writer)
        self.generate_network_source(self.writer, compiler)
        self.generate_synapses_classes_source(self.writer)
        self.generate_run_source(self.writer)
        self.copy_source_files(self.writer, directory)

        self.writer.source_files.update(additional_source_files)

        self.generate_makefile(
            self.writer,
            compiler,
            compiler_flags=" ".join(compiler_flags),
            linker_flags=" ".join(linker_flags),
            nb_threads=nb_threads,
            debug=debug,
        )

        if compile:
            self.compile_source(directory, compiler, debug, clean)
            if run:
                self.run(directory, results_directory, with_output, run_args)
        time_measurements = {
            "'make clean'": self.timers["compile"]["clean"],
            "'make'": self.timers["compile"]["make"],
            "running 'main'": self.timers["run_binary"],
        }
        logged_times = [
            f"{task}: {measurement:.2f}s"
            for task, measurement in time_measurements.items()
            if measurement is not None
        ]
        logger.debug(f"Time measurements: {', '.join(logged_times)}")

    def delete(self, code=True, data=True, run_args=True, directory=True, force=False):
        if self.project_dir is None:
            return  # Nothing to delete

        if directory and not all([code, data, run_args]):
            raise ValueError(
                "When deleting the directory, code and data will"
                "be deleted as well. Set the corresponding "
                "parameters to True."
            )

        fnames = []

        # Delete data
        if data:
            results_dir = self.results_dir
            logger.debug(f"Deleting data files in '{results_dir}'")
            fnames.append(os.path.join(results_dir, "last_run_info.txt"))
            if self.profiled_codeobjects:
                fnames.append(os.path.join(results_dir, "profiling_info.txt"))
            for var in self.arrays:
                fnames.append(os.path.join(results_dir, self.get_array_filename(var)))

        # Delete code
        if code:
            logger.debug(f"Deleting code files in '{self.project_dir}'")
            if sys.platform == "win32":
                fnames.extend(
                    [
                        "sourcefiles.txt",
                        "win_makefile",
                        "main.exe",
                        "main.ilk",
                        "main.pdb",
                        "winmake.log",
                    ]
                )
            else:
                fnames.extend(["make.deps", "makefile", "main"])

            fnames.extend(
                [
                    os.path.join("brianlib", "spikequeue.h"),
                    os.path.join("brianlib", "stdint_compat.h"),
                ]
            )
            fnames.extend(self.writer.header_files)

            for source_file in self.writer.source_files:
                fnames.append(source_file)
                base_name, _ = os.path.splitext(source_file)
                if sys.platform == "win32":
                    fnames.append(f"{base_name}.obj")
                else:
                    fnames.append(f"{base_name}.o")

            for static_array_name in self.static_arrays:
                fnames.append(os.path.join("static_arrays", static_array_name))

        if run_args:
            for fname in self.run_args_arrays:
                fnames.append(os.path.join("static_arrays", fname))

        for fname in fnames:
            full_fname = os.path.join(self.project_dir, fname)
            try:
                os.remove(full_fname)
            except OSError as ex:
                logger.warn(f'File "{full_fname}" could not be deleted: {str(ex)}')

        # Delete directories

        if directory:
            directories = [
                "brianlib",
                "code_objects",
                "results",
                "static_arrays",
                "",
            ]
            full_directories = [
                os.path.join(self.project_dir, directory) for directory in directories
            ]
            for full_directory in full_directories:
                try:
                    os.rmdir(full_directory)
                except OSError:
                    if not os.path.exists(full_directory):
                        continue

                    # The directory is not empty:
                    if force:
                        logger.debug(
                            'Directory "{}" is not empty, but '
                            "deleting it due to the use of the force "
                            "option."
                        )
                        shutil.rmtree(full_directory)
                    else:
                        # We only give a warning if there is a file or a
                        # directory we do not know about. We do not want to e.g.
                        # complain about an unknown file in the results
                        # directory and then again complain about the results
                        # directory when deleting the main directory
                        still_present = [
                            name
                            for name in os.listdir(full_directory)
                            if os.path.isfile(name)
                            or os.path.join(full_directory, name)
                            not in full_directories
                        ]
                        if len(still_present):
                            still_present = ", ".join(
                                f'"{name}"' for name in still_present
                            )
                            logger.warn(
                                f"Not deleting the '{full_directory}' directory, "
                                "because it contains files/directories "
                                f"not added by Brian: {still_present}",
                                name_suffix="delete_skips_directory",
                            )

    def network_run(
        self,
        net,
        duration,
        report=None,
        report_period=10 * second,
        namespace=None,
        profile=None,
        level=0,
        **kwds,
    ):
        if duration < 0:
            raise ValueError(
                f"Function 'run' expected a non-negative duration but got '{duration}'"
            )

        self.networks.add(net)
        if kwds:
            logger.warn(
                "Unsupported keyword argument(s) provided for run: %s"
                % ", ".join(kwds.keys())
            )
        # We store this as an instance variable for later access by the
        # `code_object` method
        self.enable_profiling = profile

        # Allow setting `profile` in the `set_device` call (used e.g. in brian2cuda
        # SpeedTest configurations)
        if profile is None:
            self.enable_profiling = self.build_options.get("profile", False)

        all_objects = net.sorted_objects
        net._clocks = {obj.clock for obj in all_objects}
        t_end = net.t + duration
        for clock in net._clocks:
            clock.set_interval(net.t, t_end)

        # Get the local namespace
        if namespace is None:
            namespace = get_local_namespace(level=level + 2)

        net.before_run(namespace)
        self.synapses |= {s for s in net.objects if isinstance(s, Synapses)}
        self.clocks.update(net._clocks)
        net.t_ = float(t_end)

        # TODO: remove this horrible hack
        for clock in self.clocks:
            if clock.name == "clock":
                clock._name = "_clock"

        # Extract all the CodeObjects
        # Note that since we ran the Network object, these CodeObjects will be sorted into the right
        # running order, assuming that there is only one clock
        code_objects = []
        for obj in all_objects:
            if obj.active:
                for codeobj in obj._code_objects:
                    code_objects.append((obj.clock, codeobj))

        # Code for a progress reporting function
        standard_code = """
        std::string _format_time(float time_in_s)
        {
            float divisors[] = {24*60*60, 60*60, 60, 1};
            char letters[] = {'d', 'h', 'm', 's'};
            float remaining = time_in_s;
            std::string text = "";
            int time_to_represent;
            for (int i =0; i < sizeof(divisors)/sizeof(float); i++)
            {
                time_to_represent = int(remaining / divisors[i]);
                remaining -= time_to_represent * divisors[i];
                if (time_to_represent > 0 || text.length())
                {
                    if(text.length() > 0)
                    {
                        text += " ";
                    }
                    text += (std::to_string(time_to_represent)+letters[i]);
                }
            }
            //less than one second
            if(text.length() == 0)
            {
                text = "< 1s";
            }
            return text;
        }
        void report_progress(const double elapsed, const double completed, const double start, const double duration)
        {
            if (completed == 0.0)
            {
                %STREAMNAME% << "Starting simulation at t=" << start << " s for duration " << duration << " s";
            } else
            {
                %STREAMNAME% << completed*duration << " s (" << (int)(completed*100.) << "%) simulated in " << _format_time(elapsed);
                if (completed < 1.0)
                {
                    const int remaining = (int)((1-completed)/completed*elapsed+0.5);
                    %STREAMNAME% << ", estimated " << _format_time(remaining) << " remaining.";
                }
            }

            %STREAMNAME% << std::endl << std::flush;
        }
        """
        if report is None:
            report_func = ""
        elif report == "text" or report == "stdout":
            report_func = standard_code.replace("%STREAMNAME%", "std::cout")
        elif report == "stderr":
            report_func = standard_code.replace("%STREAMNAME%", "std::cerr")
        elif isinstance(report, str):
            report_func = """
            void report_progress(const double elapsed, const double completed, const double start, const double duration)
            {
            %REPORT%
            }
            """.replace(
                "%REPORT%", report
            )
        else:
            raise TypeError(
                "report argument has to be either 'text', "
                "'stdout', 'stderr', or the code for a report "
                "function"
            )

        if report_func != "":
            if self.report_func != "" and report_func != self.report_func:
                raise NotImplementedError(
                    "The C++ standalone device does not "
                    "support multiple report functions, "
                    "each run has to use the same (or "
                    "none)."
                )
            self.report_func = report_func

        if report is not None:
            report_call = "report_progress"
        else:
            report_call = "NULL"

        # Generate the updaters
        run_lines = [f"{net.name}.clear();"]
        all_clocks = set()
        for clock, codeobj in code_objects:
            run_lines.append(f"{net.name}.add(&{clock.name}, _run_{codeobj.name});")
            all_clocks.add(clock)

        # Under some rare circumstances (e.g. a NeuronGroup only defining a
        # subexpression that is used by other groups (via linking, or recorded
        # by a StateMonitor) *and* not calculating anything itself *and* using a
        # different clock than all other objects) a clock that is not used by
        # any code object should nevertheless advance during the run. We include
        # such clocks without a code function in the network.
        for clock in net._clocks:
            if clock not in all_clocks:
                run_lines.append(f"{net.name}.add(&{clock.name}, NULL);")

        run_lines.extend(self.code_lines["before_network_run"])
        if not self.run_args_applied:
            run_lines.append("set_from_command_line(args);")
            self.run_args_applied = True
        run_lines.append(
            f"{net.name}.run({float(duration)!r}, {report_call},"
            f" {float(report_period)!r});"
        )
        run_lines.extend(self.code_lines["after_network_run"])
        self.main_queue.append(("run_network", (net, run_lines)))

        net.after_run()

        # Manually set the cache for the clocks, simulation scripts might
        # want to access the time (which has been set in code and is therefore
        # not accessible by the normal means until the code has been built and
        # run)
        for clock in net._clocks:
            self.array_cache[clock.variables["timestep"]] = np.array([clock._i_end])
            self.array_cache[clock.variables["t"]] = np.array(
                [clock._i_end * clock.dt_]
            )

        if self.build_on_run:
            if self.has_been_run:
                raise RuntimeError(
                    "The network has already been built and run "
                    "before. Use set_device with "
                    "build_on_run=False and an explicit "
                    "device.build call to use multiple run "
                    "statements with this device."
                )
            self.build(direct_call=False, **self.build_options)

    def network_store(self, net, *args, **kwds):
        raise NotImplementedError(
            "The store/restore mechanism is not supported in the C++ standalone"
        )

    def network_restore(self, net, *args, **kwds):
        raise NotImplementedError(
            "The store/restore mechanism is not supported in the C++ standalone"
        )

    def network_get_profiling_info(self, net):
        fname = os.path.join(self.project_dir, "results", "profiling_info.txt")
        if not os.path.exists(fname):
            raise ValueError(
                "No profiling info collected (did you run with 'profile=True'?)"
            )
        net._profiling_info = []
        with open(fname) as f:
            for line in f:
                (key, val) = line.split()
                net._profiling_info.append((key, float(val) * second))
        return sorted(net._profiling_info, key=lambda item: item[1], reverse=True)

    def run_function(self, name, include_in_parent=True):
        """
        Context manager to divert code into a function

        Code that happens within the scope of this context manager will go into the named function.

        Parameters
        ----------
        name : str
            The name of the function to divert code into.
        include_in_parent : bool
            Whether or not to include a call to the newly defined function in the parent context.
        """
        return RunFunctionContext(name, include_in_parent)


class RunFunctionContext:
    def __init__(self, name, include_in_parent):
        self.name = name
        self.include_in_parent = include_in_parent

    def __enter__(self):
        cpp_standalone_device.main_queue.append(
            ("start_run_func", (self.name, self.include_in_parent))
        )

    def __exit__(self, type, value, traceback):
        cpp_standalone_device.main_queue.append(
            ("end_run_func", (self.name, self.include_in_parent))
        )


cpp_standalone_device = CPPStandaloneDevice()
all_devices["cpp_standalone"] = cpp_standalone_device



#!/usr/bin/env python
"""
Distributed synchrony example
=============================
Fig. 14 from:

    Rossant C, Leijon S, Magnusson AK, Brette R (2011).
    "Sensitivity of noisy neurons to coincident inputs".
    Journal of Neuroscience, 31(47).

5000 independent E/I Poisson inputs are injected into a leaky integrate-and-fire neuron.
Synchronous events, following an independent Poisson process at 40 Hz, are considered, 
where 15 E Poisson spikes are randomly shifted to be synchronous at those events.
The output firing rate is then significantly higher, showing that the spike timing of
less than 1% of the excitatory synapses have an important impact on the postsynaptic firing.
"""
from brian2 import *

# neuron parameters
theta = -55*mV
El = -65*mV
vmean = -65*mV
taum = 5*ms
taue = 3*ms
taui = 10*ms
eqs = Equations("""
                dv/dt  = (ge+gi-(v-El))/taum : volt
                dge/dt = -ge/taue : volt
                dgi/dt = -gi/taui : volt
                """)

# input parameters
p = 15
ne = 4000
ni = 1000
lambdac = 40*Hz
lambdae = lambdai = 1*Hz

# synapse parameters
we = .5*mV/(taum/taue)**(taum/(taue-taum))
wi = (vmean-El-lambdae*ne*we*taue)/(lambdae*ni*taui)

# NeuronGroup definition
group = NeuronGroup(N=2, model=eqs, reset='v = El',
                    threshold='v>theta',
                    refractory=5*ms, method='exact')
group.v = El
group.ge = group.gi = 0

# independent E/I Poisson inputs
p1 = PoissonInput(group[0:1], 'ge', N=ne, rate=lambdae, weight=we)
p2 = PoissonInput(group[0:1], 'gi', N=ni, rate=lambdai, weight=wi)

# independent E/I Poisson inputs + synchronous E events
p3 = PoissonInput(group[1:], 'ge', N=ne, rate=lambdae-(p*1.0/ne)*lambdac, weight=we)
p4 = PoissonInput(group[1:], 'gi', N=ni, rate=lambdai, weight=wi)
p5 = PoissonInput(group[1:], 'ge', N=1, rate=lambdac, weight=p*we)

# run the simulation
M = SpikeMonitor(group)
SM = StateMonitor(group, 'v', record=True)
BrianLogger.log_level_info()
run(1*second)
# plot trace and spikes
for i in [0, 1]:
    spikes = (M.t[M.i == i] - defaultclock.dt)/ms
    val = SM[i].v
    subplot(2, 1, i+1)
    plot(SM.t/ms, val)
    plot(tile(spikes, (2, 1)), 
         vstack((val[array(spikes, dtype=int)],
                 zeros(len(spikes)))), 'C0')
    title("%s: %d spikes/second" % (["uncorrelated inputs", "correlated inputs"][i], 
                                    M.count[i]))
tight_layout()
show()



"""
Check that the features of `Synapses` are available and correct.
"""

import numpy

from brian2 import *
from brian2.tests.features import FeatureTest, InaccuracyError


class SynapsesPre(FeatureTest):
    category = "Synapses"
    name = "Presynaptic code"
    tags = ["NeuronGroup", "run", "Synapses", "Presynaptic code"]

    def run(self):
        tau = 5 * ms
        eqs = """
        dV/dt = k/tau : 1
        k : 1
        """
        G = NeuronGroup(10, eqs, threshold="V>1", reset="V=0")
        G.k = linspace(1, 5, len(G))
        H = NeuronGroup(10, "V:1")
        S = Synapses(G, H, on_pre="V += 1")
        S.connect(j="i")
        self.H = H
        run(101 * ms)

    def results(self):
        return self.H.V[:]

    compare = FeatureTest.compare_arrays


class SynapsesPost(FeatureTest):
    category = "Synapses"
    name = "Postsynaptic code"
    tags = ["NeuronGroup", "run", "Synapses", "Postsynaptic code"]

    def run(self):
        tau = 5 * ms
        eqs = """
        dV/dt = k/tau : 1
        k : 1
        """
        G = NeuronGroup(10, eqs, threshold="V>1", reset="V=0")
        G.k = linspace(1, 5, len(G))
        H = NeuronGroup(10, "V:1")
        S = Synapses(H, G, on_post="V_pre += 1")
        S.connect(j="i")
        self.H = H
        run(101 * ms)

    def results(self):
        return self.H.V[:]

    compare = FeatureTest.compare_arrays


class SynapsesSTDP(FeatureTest):
    category = "Synapses"
    name = "STDP"
    tags = [
        "NeuronGroup",
        "Threshold",
        "Reset",
        "Refractory",
        "run",
        "Synapses",
        "Postsynaptic code",
        "Presynaptic code",
        "SpikeMonitor",
        "StateMonitor",
        "SpikeGeneratorGroup",
    ]

    def run(self):
        n_cells = 100
        n_recorded = 10
        numpy.random.seed(42)
        taum = 20 * ms
        taus = 5 * ms
        Vt = -50 * mV
        Vr = -60 * mV
        El = -49 * mV
        fac = 60 * 0.27 / 10
        gmax = 20 * fac
        dApre = 0.01
        taupre = 20 * ms
        taupost = taupre
        dApost = -dApre * taupre / taupost * 1.05
        dApost *= 0.1 * gmax
        dApre *= 0.1 * gmax

        connectivity = numpy.random.randn(n_cells, n_cells)
        sources = numpy.random.random_integers(0, n_cells - 1, 10 * n_cells)
        # Only use one spike per time step (to rule out that a single source neuron
        # has more than one spike in a time step)
        times = (
            numpy.random.choice(numpy.arange(10 * n_cells), 10 * n_cells, replace=False)
            * ms
        )
        v_init = Vr + numpy.random.rand(n_cells) * (Vt - Vr)

        eqs = Equations(
            """
            dv/dt = (g-(v-El))/taum : volt
            dg/dt = -g/taus         : volt
            """
        )

        P = NeuronGroup(
            n_cells, model=eqs, threshold="v>Vt", reset="v=Vr", refractory=5 * ms
        )
        Q = SpikeGeneratorGroup(n_cells, sources, times)
        P.v = v_init
        P.g = 0 * mV
        S = Synapses(
            P,
            P,
            model="""dApre/dt=-Apre/taupre    : 1 (event-driven)    
                     dApost/dt=-Apost/taupost : 1 (event-driven)
                     w                        : 1""",
            pre="""g     += w*mV
                   Apre  += dApre
                   w      = w + Apost""",
            post="""Apost += dApost
                    w      = w + Apre""",
        )
        S.connect()

        S.w = fac * connectivity.flatten()

        T = Synapses(Q, P, model="w : 1", on_pre="g += w*mV")
        T.connect(j="i")
        T.w = 10 * fac

        spike_mon = SpikeMonitor(P)
        state_mon = StateMonitor(S, "w", record=np.arange(n_recorded))
        v_mon = StateMonitor(P, "v", record=np.arange(n_recorded))

        self.state_mon = state_mon
        self.spike_mon = spike_mon
        self.v_mon = v_mon

        run(0.2 * second, report="text")

    def results(self):
        return self.state_mon.w[:], self.v_mon.v[:], self.spike_mon.num_spikes

    def compare(self, maxrelerr, res1, res2):
        w1, v1, n1 = res1
        w2, v2, n2 = res2
        FeatureTest.compare_arrays(self, maxrelerr, w1, w2)
        FeatureTest.compare_arrays(self, maxrelerr, v1, v2)
        FeatureTest.compare_arrays(
            self, maxrelerr, array([n1], dtype=float), array([n2], dtype=float)
        )


if __name__ == "__main__":
    for ftc in [SynapsesPre, SynapsesPost]:
        ft = ftc()
        ft.run()
        print(ft.results())
